# Copyright 2010-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import collections
import io
import logging
import warnings
import re
import typing

import portage
from pathlib import Path
from portage import eclass_cache, os
from portage.checksum import get_valid_checksum_keys
from portage.const import PORTAGE_BASE_PATH, REPO_NAME_LOC, USER_CONFIG_PATH
from portage.eapi import (
    eapi_allows_directories_on_profile_level_and_repository_level,
    eapi_has_repo_deps,
)
from portage.env.loaders import KeyValuePairFileLoader
from portage.util import (
    normalize_path,
    read_corresponding_eapi_file,
    shlex_split,
    stack_lists,
    writemsg,
    writemsg_level,
    _recursive_file_list,
)
from portage.util.configparser import SafeConfigParser, ConfigParserError, read_configs
from portage.util._path import isdir_raise_eaccess
from portage.util.path import first_existing
from portage.localization import _
from portage import _unicode_decode
from portage import _unicode_encode
from portage import _encodings
from portage import manifest
import portage.sync

_profile_node = collections.namedtuple(
    "_profile_node",
    (
        "location",
        "portage1_directories",
        "user_config",
        "profile_formats",
        "eapi",
        "allow_build_id",
        "show_deprecated_warning",
    ),
)

# Characters prohibited by repoman's file.name check.
_invalid_path_char_re = re.compile(r"[^a-zA-Z0-9._\-+/]")

_valid_profile_formats = frozenset(
    [
        "pms",
        "portage-1",
        "portage-2",
        "profile-bashrcs",
        "profile-set",
        "profile-default-eapi",
        "build-id",
        "profile-repo-deps",
    ]
)

_portage1_profiles_allow_directories = frozenset(
    ["portage-1-compat", "portage-1", "portage-2"]
)

_repo_name_sub_re = re.compile(r"[^\w-]")


def _gen_valid_repo(name):
    """
    Substitute hyphen in place of characters that don't conform to PMS 3.1.5,
    and strip hyphen from left side if necessary. This returns None if the
    given name contains no valid characters.
    """
    name = _repo_name_sub_re.sub(" ", name.strip())
    name = "-".join(name.split())
    name = name.lstrip("-")
    if not name:
        name = None
    return name


def _find_invalid_path_char(path, pos=0, endpos=None):
    """
    Returns the position of the first invalid character found in basename,
    or -1 if no invalid characters are found.
    """
    if endpos is None:
        endpos = len(path)

    m = _invalid_path_char_re.search(path, pos=pos, endpos=endpos)
    if m is not None:
        return m.start()

    return -1


class RepoConfig:
    """Stores config of one repository"""

    __slots__ = (
        "aliases",
        "allow_missing_manifest",
        "allow_provide_virtual",
        "auto_sync",
        "cache_formats",
        "clone_depth",
        "create_manifest",
        "disable_manifest",
        "eapi",
        "eclass_db",
        "eclass_locations",
        "eclass_overrides",
        "find_invalid_path_char",
        "force",
        "format",
        "local_config",
        "location",
        "main_repo",
        "manifest_hashes",
        "manifest_required_hashes",
        "masters",
        "missing_repo_name",
        "module_specific_options",
        "name",
        "portage1_profiles",
        "portage1_profiles_compat",
        "priority",
        "profile_formats",
        "properties_allowed",
        "restrict_allowed",
        "sign_commit",
        "sign_manifest",
        "strict_misc_digests",
        "sync_allow_hardlinks",
        "sync_depth",
        "sync_hooks_only_on_change",
        "sync_openpgp_keyserver",
        "sync_openpgp_key_path",
        "sync_openpgp_key_refresh",
        "sync_openpgp_key_refresh_retry_count",
        "sync_openpgp_key_refresh_retry_delay_exp_base",
        "sync_openpgp_key_refresh_retry_delay_max",
        "sync_openpgp_key_refresh_retry_delay_mult",
        "sync_openpgp_key_refresh_retry_overall_timeout",
        "sync_rcu",
        "sync_rcu_spare_snapshots",
        "sync_rcu_store_dir",
        "sync_rcu_ttl_days",
        "sync_type",
        "sync_umask",
        "sync_uri",
        "sync_user",
        "thin_manifest",
        "update_changelog",
        "user_location",
        "volatile",
        "_eapis_banned",
        "_eapis_deprecated",
        "_masters_orig",
    )

    def __init__(self, name, repo_opts, local_config=True):
        """Build a RepoConfig with options in repo_opts
        Try to read repo_name in repository location, but if
        it is not found use variable name as repository name"""

        force = repo_opts.get("force")
        if force is not None:
            force = tuple(force.split())
        self.force = force
        if force is None:
            force = ()

        self.local_config = local_config

        if local_config or "aliases" in force:
            aliases = repo_opts.get("aliases")
            if aliases is not None:
                aliases = tuple(aliases.split())
        else:
            aliases = None

        self.aliases = aliases

        if local_config or "eclass-overrides" in force:
            eclass_overrides = repo_opts.get("eclass-overrides")
            if eclass_overrides is not None:
                eclass_overrides = tuple(eclass_overrides.split())
        else:
            eclass_overrides = None

        self.eclass_overrides = eclass_overrides
        # Eclass databases and locations are computed later.
        self.eclass_db = None
        self.eclass_locations = None

        if local_config or "masters" in force:
            # Masters from repos.conf override layout.conf.
            masters = repo_opts.get("masters")
            if masters is not None:
                masters = tuple(masters.split())
        else:
            masters = None

        self.masters = masters

        # The main-repo key makes only sense for the 'DEFAULT' section.
        self.main_repo = repo_opts.get("main-repo")

        priority = repo_opts.get("priority")
        if priority is not None:
            try:
                priority = int(priority)
            except ValueError:
                priority = None
        self.priority = priority

        sync_type = repo_opts.get("sync-type")
        if sync_type is not None:
            sync_type = sync_type.strip()
        self.sync_type = sync_type or None

        sync_umask = repo_opts.get("sync-umask")
        if sync_umask is not None:
            sync_umask = sync_umask.strip()
        self.sync_umask = sync_umask or None

        sync_uri = repo_opts.get("sync-uri")
        if sync_uri is not None:
            sync_uri = sync_uri.strip()
        self.sync_uri = sync_uri or None

        sync_user = repo_opts.get("sync-user")
        if sync_user is not None:
            sync_user = sync_user.strip()
        self.sync_user = sync_user or None

        auto_sync = repo_opts.get("auto-sync", "yes")
        if auto_sync is not None:
            auto_sync = auto_sync.strip().lower()
        self.auto_sync = auto_sync

        self.clone_depth = repo_opts.get("clone-depth")
        self.sync_depth = repo_opts.get("sync-depth")

        self.sync_hooks_only_on_change = repo_opts.get(
            "sync-hooks-only-on-change", "false"
        ).lower() in ("true", "yes")

        self.strict_misc_digests = repo_opts.get(
            "strict-misc-digests", "true"
        ).lower() in ("true", "yes")

        self.sync_allow_hardlinks = repo_opts.get(
            "sync-allow-hardlinks", "true"
        ).lower() in ("true", "yes")

        self.sync_openpgp_keyserver = (
            repo_opts.get("sync-openpgp-keyserver", "").strip().lower() or None
        )

        self.sync_openpgp_key_path = repo_opts.get("sync-openpgp-key-path", None)

        self.sync_openpgp_key_refresh = repo_opts.get(
            "sync-openpgp-key-refresh", "true"
        ).lower() in ("true", "yes")

        for k in (
            "sync_openpgp_key_refresh_retry_count",
            "sync_openpgp_key_refresh_retry_delay_exp_base",
            "sync_openpgp_key_refresh_retry_delay_max",
            "sync_openpgp_key_refresh_retry_delay_mult",
            "sync_openpgp_key_refresh_retry_overall_timeout",
        ):
            setattr(self, k, repo_opts.get(k.replace("_", "-"), None))

        self.sync_rcu = repo_opts.get("sync-rcu", "false").lower() in ("true", "yes")

        self.sync_rcu_store_dir = repo_opts.get("sync-rcu-store-dir")

        for k in ("sync-rcu-spare-snapshots", "sync-rcu-ttl-days"):
            v = repo_opts.get(k, "").strip() or None
            if v:
                try:
                    v = int(v)
                except (OverflowError, ValueError):
                    writemsg(
                        _("!!! Invalid %s setting for repo" " %s: %s\n") % (k, name, v),
                        noiselevel=-1,
                    )
                    v = None
            setattr(self, k.replace("-", "_"), v)

        self.module_specific_options = {}

        # Not implemented.
        repo_format = repo_opts.get("format")
        if repo_format is not None:
            repo_format = repo_format.strip()
        self.format = repo_format

        self.user_location = None
        location = repo_opts.get("location")
        if location is not None and location.strip():
            if os.path.isdir(location) or portage._sync_mode:
                # The user_location is required for sync-rcu support,
                # since it manages a symlink which resides at that
                # location (and realpath is irreversible).
                self.user_location = location
                location = os.path.realpath(location)
        else:
            location = None
        self.location = location

        missing = True
        self.name = name
        if self.location is not None:
            self.name, missing = self._read_valid_repo_name(self.location)
            if missing:
                # The name from repos.conf has to be used here for
                # things like emerge-webrsync to work when the repo
                # is empty (bug #484950).
                if name is not None:
                    self.name = name
                if portage._sync_mode:
                    missing = False
        elif name == "DEFAULT":
            missing = False

        volatile = repo_opts.get("volatile")
        # If volatile is explicitly set, go with it.
        if volatile is not None:
            self.volatile = volatile in ("true", "yes")
        else:
            # If it's unset, we default to no (i.e. the repository is not volatile),
            # but with a heuristic for when a repository is not likely to be suitable
            # (likely to contain custom user changes).
            try:
                # If the repository doesn't exist, we can't check its ownership,
                # so err on the safe side.
                if missing or not self.location:
                    self.volatile = True
                # On Prefix, you can't rely on the ownership as a proxy for user
                # owned because the user typically owns everything.
                # But we can't access if we're on Prefix here, so use whether
                # we're under /var/db/repos instead.
                elif not self.location.startswith("/var/db/repos"):
                    self.volatile = True
                # If the owner of the repository isn't root or Portage, it's
                # an indication the user may expect to be able to safely make
                # changes in the directory, so default to volatile.
                elif Path(self.location).owner() not in ("root", "portage"):
                    self.volatile = True
                else:
                    self.volatile = False
            except Exception:
                # There's too many conditions here to refine the exception list:
                # - We lack permissions to poke at the directory (PermissionError)
                # - Its UID doesn't actually exist and the repository
                #   won't be synced by the user (KeyError).
                # - The directory doesn't exist (FileNotFoundError)
                # - Probably many others.
                # So, just fail safe.
                self.volatile = True

        self.eapi = None
        self.missing_repo_name = missing
        # sign_commit is disabled by default, since it requires Git >=1.7.9,
        # and key_id configured by `git config user.signingkey key_id`
        self.sign_commit = False
        self.sign_manifest = True
        self.thin_manifest = False
        self.allow_missing_manifest = False
        self.allow_provide_virtual = False
        self.create_manifest = True
        self.disable_manifest = False
        self.manifest_hashes = None
        self.manifest_required_hashes = None
        self.update_changelog = False
        self.cache_formats = None
        self.portage1_profiles = True
        self.portage1_profiles_compat = False
        self.find_invalid_path_char = _find_invalid_path_char
        self._masters_orig = None

        # Parse layout.conf.
        if self.location:
            layout_data = parse_layout_conf(self.location, self.name)[0]
            self._masters_orig = layout_data["masters"]

            # layout.conf masters may be overridden here if we have a masters
            # setting from the user's repos.conf
            if self.masters is None:
                self.masters = layout_data["masters"]

            if (local_config or "aliases" in force) and layout_data["aliases"]:
                aliases = self.aliases
                if aliases is None:
                    aliases = ()
                # repos.conf aliases come after layout.conf aliases, giving
                # them the ability to do incremental overrides
                self.aliases = layout_data["aliases"] + tuple(aliases)

            if layout_data["repo-name"]:
                # allow layout.conf to override repository name
                # useful when having two copies of the same repo enabled
                # to avoid modifying profiles/repo_name in one of them
                self.name = layout_data["repo-name"]
                self.missing_repo_name = False

            for value in (
                "allow-missing-manifest",
                "cache-formats",
                "create-manifest",
                "disable-manifest",
                "manifest-hashes",
                "manifest-required-hashes",
                "profile-formats",
                "properties-allowed",
                "restrict-allowed",
                "sign-commit",
                "sign-manifest",
                "thin-manifest",
                "update-changelog",
            ):
                setattr(self, value.lower().replace("-", "_"), layout_data[value])

            # If profile-formats specifies a default EAPI, then set
            # self.eapi to that, otherwise set it to "0" as specified
            # by PMS.
            self.eapi = layout_data.get("profile_eapi_when_unspecified", "0")

            eapi = read_corresponding_eapi_file(
                os.path.join(self.location, REPO_NAME_LOC), default=self.eapi
            )

            self.portage1_profiles = (
                eapi_allows_directories_on_profile_level_and_repository_level(eapi)
                or any(
                    x in _portage1_profiles_allow_directories
                    for x in layout_data["profile-formats"]
                )
            )
            self.portage1_profiles_compat = (
                not eapi_allows_directories_on_profile_level_and_repository_level(eapi)
                and layout_data["profile-formats"] == ("portage-1-compat",)
            )

            self._eapis_banned = frozenset(layout_data["eapis-banned"])
            self._eapis_deprecated = frozenset(layout_data["eapis-deprecated"])

    def set_module_specific_opt(self, opt, val):
        self.module_specific_options[opt] = val

    def eapi_is_banned(self, eapi):
        return eapi in self._eapis_banned

    def eapi_is_deprecated(self, eapi):
        return eapi in self._eapis_deprecated

    def iter_pregenerated_caches(self, auxdbkeys, readonly=True, force=False):
        """
        Reads layout.conf cache-formats from left to right and yields cache
        instances for each supported type that's found. If no cache-formats
        are specified in layout.conf, 'pms' type is assumed if the
        metadata/cache directory exists or force is True.
        """
        formats = self.cache_formats
        if not formats:
            if not force:
                return
            # The default egencache format was 'pms' prior to portage-2.1.11.32
            # (portage versions prior to portage-2.1.11.14 will NOT
            # recognize md5-dict format unless it is explicitly listed in
            # layout.conf).
            formats = ("md5-dict",)

        for fmt in formats:
            name = None
            if fmt == "pms":
                from portage.cache.metadata import database

                name = "metadata/cache"
            elif fmt == "md5-dict":
                from portage.cache.flat_hash import md5_database as database

                name = "metadata/md5-cache"

            if name is not None:
                yield database(self.location, name, auxdbkeys, readonly=readonly)

    def get_pregenerated_cache(self, auxdbkeys, readonly=True, force=False):
        """
        Returns the first cache instance yielded from
        iter_pregenerated_caches(), or None if no cache is available or none
        of the available formats are supported.
        """
        return next(
            self.iter_pregenerated_caches(auxdbkeys, readonly=readonly, force=force),
            None,
        )

    def load_manifest(self, *args, **kwds):
        kwds["thin"] = self.thin_manifest
        kwds["allow_missing"] = self.allow_missing_manifest
        kwds["allow_create"] = self.create_manifest
        kwds["hashes"] = self.manifest_hashes
        kwds["required_hashes"] = self.manifest_required_hashes
        kwds["strict_misc_digests"] = self.strict_misc_digests
        if self.disable_manifest:
            kwds["from_scratch"] = True
        kwds["find_invalid_path_char"] = self.find_invalid_path_char
        return manifest.Manifest(*args, **kwds)

    def update(self, new_repo):
        """Update repository with options in another RepoConfig"""

        keys = set(self.__slots__)
        keys.discard("missing_repo_name")
        for k in keys:
            v = getattr(new_repo, k, None)
            if v is not None:
                setattr(self, k, v)

        if new_repo.name is not None:
            self.missing_repo_name = new_repo.missing_repo_name

    @property
    def writable(self):
        """
        Check if self.location is writable, or permissions are sufficient
        to create it if it does not exist yet.
        @rtype: bool
        @return: True if self.location is writable or can be created,
                False otherwise
        """
        return os.access(first_existing(self.location), os.W_OK)

    @staticmethod
    def _read_valid_repo_name(repo_path):
        name, missing = RepoConfig._read_repo_name(repo_path)
        # We must ensure that the name conforms to PMS 3.1.5
        # in order to avoid InvalidAtom exceptions when we
        # use it to generate atoms.
        name = _gen_valid_repo(name)
        if not name:
            # name only contains invalid characters
            name = "x-" + os.path.basename(repo_path)
            name = _gen_valid_repo(name)
            # If basename only contains whitespace then the
            # end result is name = 'x-'.
        return name, missing

    @staticmethod
    def _read_repo_name(repo_path):
        """
        Read repo_name from repo_path.
        Returns repo_name, missing.
        """
        repo_name_path = os.path.join(repo_path, REPO_NAME_LOC)
        f = None
        try:
            f = open(
                _unicode_encode(
                    repo_name_path, encoding=_encodings["fs"], errors="strict"
                ),
                encoding=_encodings["repo.content"],
                errors="replace",
            )
            return f.readline().strip(), False
        except OSError:
            return "x-" + os.path.basename(repo_path), True
        finally:
            if f is not None:
                f.close()

    def info_string(self):
        """
        Returns a formatted string containing information about the repository.
        Used by emerge --info.
        """
        indent = " " * 4
        repo_msg = []
        repo_msg.append(self.name)
        if self.format:
            repo_msg.append(indent + "format: " + self.format)
        if self.location:
            repo_msg.append(indent + "location: " + self.location)
        if not self.strict_misc_digests:
            repo_msg.append(indent + "strict-misc-digests: false")
        if not self.sync_openpgp_key_refresh:
            repo_msg.append(indent + "sync-openpgp-key-refresh: no")
        if self.sync_type:
            repo_msg.append(indent + "sync-type: " + self.sync_type)
        if self.sync_umask:
            repo_msg.append(indent + "sync-umask: " + self.sync_umask)
        if self.sync_uri:
            repo_msg.append(indent + "sync-uri: " + self.sync_uri)
        if self.sync_user:
            repo_msg.append(indent + "sync-user: " + self.sync_user)
        if self.masters:
            repo_msg.append(
                indent + "masters: " + " ".join(master.name for master in self.masters)
            )
        if self.priority is not None:
            repo_msg.append(indent + "priority: " + str(self.priority))
        if self.aliases:
            repo_msg.append(indent + "aliases: " + " ".join(self.aliases))
        if self.eclass_overrides:
            repo_msg.append(
                indent + "eclass-overrides: " + " ".join(self.eclass_overrides)
            )
        if self.volatile is not None:
            repo_msg.append(indent + "volatile: " + str(self.volatile))
        for o, v in self.module_specific_options.items():
            if v is not None:
                repo_msg.append(indent + o + ": " + v)
        repo_msg.append("")
        return "\n".join(repo_msg)

    def __repr__(self):
        return (
            "<portage.repository.config.RepoConfig(name={!r}, location={!r})>".format(
                self.name,
                _unicode_decode(self.location),
            )
        )

    def __str__(self):
        d = {}
        for k in self.__slots__:
            d[k] = getattr(self, k, None)
        return f"{d}"


class RepoConfigLoader:
    """Loads and store config of several repositories, loaded from PORTDIR_OVERLAY or repos.conf"""

    @staticmethod
    def _add_repositories(
        portdir, portdir_overlay, prepos, ignored_map, local_config, default_portdir
    ):
        """Add overlays in PORTDIR_OVERLAY as repositories"""
        overlays = []
        portdir_orig = None
        if portdir:
            portdir = normalize_path(portdir)
            portdir_orig = portdir
            overlays.append(portdir)
        try:
            port_ov = [normalize_path(i) for i in shlex_split(portdir_overlay)]
        except ValueError as e:
            # File "/usr/lib/python3.2/shlex.py", line 168, in read_token
            # 	raise ValueError("No closing quotation")
            writemsg(
                _("!!! Invalid PORTDIR_OVERLAY:" " %s: %s\n") % (e, portdir_overlay),
                noiselevel=-1,
            )
            port_ov = []
        overlays.extend(port_ov)
        default_repo_opts = {}
        if prepos["DEFAULT"].aliases is not None:
            default_repo_opts["aliases"] = " ".join(prepos["DEFAULT"].aliases)
        if prepos["DEFAULT"].eclass_overrides is not None:
            default_repo_opts["eclass-overrides"] = " ".join(
                prepos["DEFAULT"].eclass_overrides
            )
        if prepos["DEFAULT"].masters is not None:
            default_repo_opts["masters"] = " ".join(prepos["DEFAULT"].masters)

        if overlays:
            # We need a copy of the original repos.conf data, since we're
            # going to modify the prepos dict and some of the RepoConfig
            # objects that we put in prepos may have to be discarded if
            # they get overridden by a repository with the same name but
            # a different location. This is common with repoman, for example,
            # when temporarily overriding an rsync repo with another copy
            # of the same repo from CVS.
            repos_conf = prepos.copy()
            # overlay priority is negative because we want them to be looked before any other repo
            base_priority = 0
            for ov in overlays:
                # Ignore missing directory for 'gentoo' so that
                # first sync with emerge-webrsync is possible.
                if isdir_raise_eaccess(ov) or (base_priority == 0 and ov is portdir):
                    repo_opts = default_repo_opts.copy()
                    repo_opts["location"] = ov
                    name = prepos["DEFAULT"].main_repo if ov is portdir else None
                    repo = RepoConfig(name, repo_opts, local_config=local_config)
                    # repos_conf_opts contains options from repos.conf
                    repos_conf_opts = repos_conf.get(repo.name)
                    if repos_conf_opts is not None:
                        # Selectively copy only the attributes which
                        # repos.conf is allowed to override.
                        for k in (
                            "aliases",
                            "auto_sync",
                            "clone_depth",
                            "eclass_overrides",
                            "force",
                            "masters",
                            "module_specific_options",
                            "priority",
                            "strict_misc_digests",
                            "sync_allow_hardlinks",
                            "sync_depth",
                            "sync_hooks_only_on_change",
                            "sync_openpgp_keyserver",
                            "sync_openpgp_key_path",
                            "sync_openpgp_key_refresh",
                            "sync_openpgp_key_refresh_retry_count",
                            "sync_openpgp_key_refresh_retry_delay_exp_base",
                            "sync_openpgp_key_refresh_retry_delay_max",
                            "sync_openpgp_key_refresh_retry_delay_mult",
                            "sync_openpgp_key_refresh_retry_overall_timeout",
                            "sync_rcu",
                            "sync_rcu_spare_snapshots",
                            "sync_rcu_store_dir",
                            "sync_rcu_ttl_days",
                            "sync_type",
                            "sync_umask",
                            "sync_uri",
                            "sync_user",
                            "volatile",
                        ):
                            v = getattr(repos_conf_opts, k, None)

                            # If PORTDIR_OVERLAY is set, we have to require volatile,
                            # because it'll break changes e.g. with ebuild(1) or
                            # development in a local repository with the same repo_name.
                            if k == "volatile" and portdir_overlay:
                                v = True

                            if v is not None:
                                setattr(repo, k, v)

                    if repo.name in prepos:
                        # Silently ignore when PORTDIR overrides the location
                        # setting from the default repos.conf (bug #478544).
                        old_location = prepos[repo.name].location
                        if (
                            old_location is not None
                            and old_location != repo.location
                            and not (
                                base_priority == 0 and old_location == default_portdir
                            )
                        ):
                            ignored_map.setdefault(repo.name, []).append(old_location)
                            if old_location == portdir:
                                portdir = repo.location

                    if repo.priority is None:
                        if base_priority == 0 and ov == portdir_orig:
                            # If it's the original PORTDIR setting and it's not
                            # in PORTDIR_OVERLAY, then it will be assigned a
                            # special priority setting later.
                            pass
                        else:
                            repo.priority = base_priority
                            base_priority += 1

                    prepos[repo.name] = repo
                else:
                    if not portage._sync_mode:
                        writemsg(
                            _("!!! Invalid PORTDIR_OVERLAY (not a dir): '%s'\n") % ov,
                            noiselevel=-1,
                        )

        return portdir

    @staticmethod
    def _parse(paths, prepos, local_config, default_opts):
        """Parse files in paths to load config"""
        parser = SafeConfigParser(defaults=default_opts)

        recursive_paths = []
        for p in paths:
            if isinstance(p, str):
                recursive_paths.extend(_recursive_file_list(p))
            else:
                recursive_paths.append(p)

        read_configs(parser, recursive_paths)

        prepos["DEFAULT"] = RepoConfig(
            "DEFAULT", parser.defaults(), local_config=local_config
        )

        for sname in parser.sections():
            optdict = {}
            for oname in parser.options(sname):
                optdict[oname] = parser.get(sname, oname)

            repo = RepoConfig(sname, optdict, local_config=local_config)
            for o in portage.sync.module_specific_options(repo):
                if parser.has_option(sname, o):
                    repo.set_module_specific_opt(o, parser.get(sname, o))

            # Perform repos.conf sync variable validation
            portage.sync.validate_config(repo, logging)

            # For backward compatibility with locations set via PORTDIR and
            # PORTDIR_OVERLAY, delay validation of the location and repo.name
            # until after PORTDIR and PORTDIR_OVERLAY have been processed.
            prepos[sname] = repo

    def __init__(self, paths, settings):
        """Load config from files in paths"""

        prepos = {}
        location_map = {}
        treemap = {}
        ignored_map = {}
        default_opts = {
            "EPREFIX": settings["EPREFIX"],
            "EROOT": settings["EROOT"],
            "PORTAGE_CONFIGROOT": settings["PORTAGE_CONFIGROOT"],
            "ROOT": settings["ROOT"],
        }

        if "PORTAGE_REPOSITORIES" in settings:
            portdir = ""
            portdir_overlay = ""
            # deprecated portdir_sync
            portdir_sync = ""
        else:
            portdir = settings.get("PORTDIR", "")
            portdir_overlay = settings.get("PORTDIR_OVERLAY", "")
            # deprecated portdir_sync
            portdir_sync = settings.get("SYNC", "")

        default_opts["sync-rsync-extra-opts"] = settings.get(
            "PORTAGE_RSYNC_EXTRA_OPTS", ""
        )

        try:
            self._parse(paths, prepos, settings.local_config, default_opts)
        except ConfigParserError as e:
            writemsg(
                _("!!! Error while reading repo config file: %s\n") % e, noiselevel=-1
            )
            # The configparser state is unreliable (prone to quirky
            # exceptions) after it has thrown an error, so use empty
            # config and try to fall back to PORTDIR{,_OVERLAY}.
            prepos.clear()
            prepos["DEFAULT"] = RepoConfig(
                "DEFAULT", {}, local_config=settings.local_config
            )
            location_map.clear()
            treemap.clear()

        repo_locations = frozenset(repo.location for repo in prepos.values())
        for repo_location in ("var/db/repos/gentoo", "usr/portage"):
            default_portdir = os.path.join(
                os.sep, settings["EPREFIX"].lstrip(os.sep), repo_location
            )
            if default_portdir in repo_locations:
                break

        # If PORTDIR_OVERLAY contains a repo with the same repo_name as
        # PORTDIR, then PORTDIR is overridden.
        portdir = self._add_repositories(
            portdir,
            portdir_overlay,
            prepos,
            ignored_map,
            settings.local_config,
            default_portdir,
        )
        if portdir and portdir.strip():
            portdir = os.path.realpath(portdir)

        ignored_repos = tuple(
            (repo_name, tuple(paths)) for repo_name, paths in ignored_map.items()
        )

        self.missing_repo_names = frozenset(
            repo.location
            for repo in prepos.values()
            if repo.location is not None and repo.missing_repo_name
        )

        # Do this before expanding aliases, so that location_map and
        # treemap consistently map unaliased names whenever available.
        for repo_name, repo in list(prepos.items()):
            if repo.location is None:
                if repo_name != "DEFAULT":
                    # Skip this warning for repoman (bug #474578).
                    if settings.local_config and paths:
                        writemsg_level(
                            "!!! %s\n"
                            % _(
                                "Section '%s' in repos.conf is missing location attribute"
                            )
                            % repo.name,
                            level=logging.ERROR,
                            noiselevel=-1,
                        )
                    del prepos[repo_name]
                    continue
            else:
                if not portage._sync_mode:
                    if not isdir_raise_eaccess(repo.location):
                        writemsg_level(
                            "!!! %s\n"
                            % _(
                                "Section '%s' in repos.conf has location attribute set "
                                "to nonexistent directory: '%s'"
                            )
                            % (repo_name, repo.location),
                            level=logging.ERROR,
                            noiselevel=-1,
                        )

                        # Ignore missing directory for 'gentoo' so that
                        # first sync with emerge-webrsync is possible.
                        if repo.name != "gentoo":
                            del prepos[repo_name]
                            continue

                    # After removing support for PORTDIR_OVERLAY, the following check can be:
                    # if repo.missing_repo_name:
                    if repo.missing_repo_name and repo.name != repo_name:
                        writemsg_level(
                            "!!! %s\n"
                            % _(
                                "Section '%s' in repos.conf refers to repository "
                                "without repository name set in '%s'"
                            )
                            % (repo_name, os.path.join(repo.location, REPO_NAME_LOC)),
                            level=logging.ERROR,
                            noiselevel=-1,
                        )
                        del prepos[repo_name]
                        continue

                    if repo.name != repo_name:
                        writemsg_level(
                            "!!! %s\n"
                            % _(
                                "Section '%s' in repos.conf has name different "
                                "from repository name '%s' set inside repository"
                            )
                            % (repo_name, repo.name),
                            level=logging.ERROR,
                            noiselevel=-1,
                        )
                        del prepos[repo_name]
                        continue

                location_map[repo.location] = repo_name
                treemap[repo_name] = repo.location

        # Add alias mappings, but never replace unaliased mappings.
        for repo_name, repo in list(prepos.items()):
            names = set()
            names.add(repo_name)
            if repo.aliases:
                aliases = stack_lists([repo.aliases], incremental=True)
                names.update(aliases)

            for name in names:
                if name in prepos and prepos[name].location is not None:
                    if name == repo_name:
                        # unaliased names already handled earlier
                        continue
                    writemsg_level(
                        _(
                            "!!! Repository name or alias '%s', "
                            + "defined for repository '%s', overrides "
                            + "existing alias or repository.\n"
                        )
                        % (name, repo_name),
                        level=logging.WARNING,
                        noiselevel=-1,
                    )
                    # Never replace an unaliased mapping with
                    # an aliased mapping.
                    continue
                prepos[name] = repo
                if repo.location is not None:
                    if repo.location not in location_map:
                        # Never replace an unaliased mapping with
                        # an aliased mapping.
                        location_map[repo.location] = name
                    treemap[name] = repo.location

        main_repo = prepos["DEFAULT"].main_repo
        if main_repo is None or main_repo not in prepos:
            # setting main_repo if it was not set in repos.conf
            main_repo = location_map.get(portdir)
            if main_repo is not None:
                prepos["DEFAULT"].main_repo = main_repo
            else:
                prepos["DEFAULT"].main_repo = None
                if portdir and not portage._sync_mode:
                    writemsg(
                        _("!!! main-repo not set in DEFAULT and PORTDIR is empty.\n"),
                        noiselevel=-1,
                    )

        if main_repo is not None and prepos[main_repo].priority is None:
            # This happens if main-repo has been set in repos.conf.
            prepos[main_repo].priority = -1000

        # DEPRECATED Backward compatible SYNC support for old mirrorselect.
        # Feb. 2, 2015.  Version 2.2.16
        if portdir_sync and main_repo is not None:
            writemsg(
                _(
                    "!!! SYNC setting found in make.conf.\n    "
                    "This setting is Deprecated and no longer used.  "
                    "Please ensure your 'sync-type' and 'sync-uri' are set correctly"
                    " in /etc/portage/repos.conf/gentoo.conf\n"
                ),
                noiselevel=-1,
            )

        # Include repo.name in sort key, for predictable sorting
        # even when priorities are equal.
        prepos_order = sorted(
            prepos.items(), key=lambda r: (r[1].priority or 0, r[1].name)
        )

        # filter duplicates from aliases, by only including
        # items where repo.name == key
        prepos_order = [
            repo.name
            for (key, repo) in prepos_order
            if repo.name == key and key != "DEFAULT" and repo.location is not None
        ]

        self.prepos = prepos
        self.prepos_order = prepos_order
        self.ignored_repos = ignored_repos
        self.location_map = location_map
        self.treemap = treemap
        self._prepos_changed = True
        self._repo_location_list = []

        # The 'masters' key currently contains repo names. Replace them with the matching RepoConfig.
        for repo_name, repo in prepos.items():
            if repo_name == "DEFAULT":
                continue
            if repo.masters is None:
                if self.mainRepo() and repo_name != self.mainRepo().name:
                    repo.masters = (self.mainRepo(),)
                else:
                    repo.masters = ()
            else:
                if repo.masters and isinstance(repo.masters[0], RepoConfig):
                    # This one has already been processed
                    # because it has an alias.
                    continue
                master_repos = []
                for master_name in repo.masters:
                    if master_name not in prepos:
                        layout_filename = os.path.join(
                            repo.location, "metadata", "layout.conf"
                        )
                        writemsg_level(
                            _(
                                "Unavailable repository '%s' "
                                "referenced by masters entry in '%s'\n"
                            )
                            % (master_name, layout_filename),
                            level=logging.ERROR,
                            noiselevel=-1,
                        )
                    else:
                        master_repos.append(prepos[master_name])
                repo.masters = tuple(master_repos)

        # The 'eclass_overrides' key currently contains repo names. Replace them with the matching repo paths.
        for repo_name, repo in prepos.items():
            if repo_name == "DEFAULT":
                continue

            eclass_locations = []
            eclass_locations.extend(
                master_repo.location for master_repo in repo.masters
            )
            # Only append the current repo to eclass_locations if it's not
            # there already. This allows masters to have more control over
            # eclass override order, which may be useful for scenarios in
            # which there is a plan to migrate eclasses to a master repo.
            if repo.location not in eclass_locations:
                eclass_locations.append(repo.location)

            if repo.eclass_overrides:
                for other_repo_name in repo.eclass_overrides:
                    if other_repo_name in self.treemap:
                        eclass_locations.append(
                            self.get_location_for_name(other_repo_name)
                        )
                    else:
                        writemsg_level(
                            _(
                                "Unavailable repository '%s' "
                                "referenced by eclass-overrides entry for "
                                "'%s'\n"
                            )
                            % (other_repo_name, repo_name),
                            level=logging.ERROR,
                            noiselevel=-1,
                        )
            repo.eclass_locations = tuple(eclass_locations)

        eclass_dbs = {}
        for repo_name, repo in prepos.items():
            if repo_name == "DEFAULT":
                continue

            eclass_db = None
            for eclass_location in repo.eclass_locations:
                tree_db = eclass_dbs.get(eclass_location)
                if tree_db is None:
                    tree_db = eclass_cache.cache(eclass_location)
                    eclass_dbs[eclass_location] = tree_db
                if eclass_db is None:
                    eclass_db = tree_db.copy()
                else:
                    eclass_db.append(tree_db)
            repo.eclass_db = eclass_db

        for repo_name, repo in prepos.items():
            if repo_name == "DEFAULT":
                continue

            if (
                repo._masters_orig is None
                and self.mainRepo()
                and repo.name != self.mainRepo().name
                and not portage._sync_mode
            ):
                # TODO: Delete masters code in lib/portage/tests/resolver/ResolverPlayground.py when deleting this warning.
                writemsg_level(
                    "!!! %s\n"
                    % _("Repository '%s' is missing masters attribute in '%s'")
                    % (
                        repo.name,
                        os.path.join(repo.location, "metadata", "layout.conf"),
                    )
                    + "!!! %s\n"
                    % _("Set 'masters = %s' in this file for future compatibility")
                    % self.mainRepo().name,
                    level=logging.WARNING,
                    noiselevel=-1,
                )

        self._prepos_changed = True
        self._repo_location_list = []

        self._check_locations()

    def repoLocationList(self):
        """Get a list of repositories location. Replaces PORTDIR_OVERLAY"""
        if self._prepos_changed:
            _repo_location_list = []
            for repo in self.prepos_order:
                if self.prepos[repo].location is not None:
                    _repo_location_list.append(self.prepos[repo].location)
            self._repo_location_list = tuple(_repo_location_list)

            self._prepos_changed = False
        return self._repo_location_list

    def mainRepoLocation(self):
        """Returns the location of main repo"""
        main_repo = self.prepos["DEFAULT"].main_repo
        if main_repo is not None and main_repo in self.prepos:
            return self.prepos[main_repo].location
        return ""

    def mainRepo(self):
        """Returns the main repo"""
        main_repo = self.prepos["DEFAULT"].main_repo
        if main_repo is None:
            return None
        return self.prepos[main_repo]

    def _check_locations(self):
        """Check if repositories location are correct and show a warning message if not"""
        for name, r in self.prepos.items():
            if name != "DEFAULT":
                if r.location is None:
                    writemsg(
                        _("!!! Location not set for repository %s\n") % name,
                        noiselevel=-1,
                    )
                else:
                    if not isdir_raise_eaccess(r.location) and not portage._sync_mode:
                        self.prepos_order.remove(name)
                        writemsg(
                            _("!!! Invalid Repository Location" " (not a dir): '%s'\n")
                            % r.location,
                            noiselevel=-1,
                        )

    def repos_with_profiles(self):
        for repo_name in self.prepos_order:
            repo = self.prepos[repo_name]
            if repo.format != "unavailable":
                yield repo

    def get_name_for_location(self, location):
        return self.location_map[location]

    def get_location_for_name(self, repo_name):
        if repo_name is None:
            # This simplifies code in places where
            # we want to be able to pass in Atom.repo
            # even if it is None.
            return None
        return self.treemap[repo_name]

    def get_repo_for_location(self, location):
        return self.prepos[self.get_name_for_location(location)]

    def __setitem__(self, repo_name, repo):
        # self.prepos[repo_name] = repo
        raise NotImplementedError

    def __getitem__(self, repo_name):
        return self.prepos[repo_name]

    def __delitem__(self, repo_name):
        if repo_name == self.prepos["DEFAULT"].main_repo:
            self.prepos["DEFAULT"].main_repo = None
        location = self.prepos[repo_name].location
        del self.prepos[repo_name]
        if repo_name in self.prepos_order:
            self.prepos_order.remove(repo_name)
        for k, v in self.location_map.copy().items():
            if v == repo_name:
                del self.location_map[k]
        if repo_name in self.treemap:
            del self.treemap[repo_name]
        self._repo_location_list = tuple(
            x for x in self._repo_location_list if x != location
        )

    def __iter__(self):
        for repo_name in self.prepos_order:
            yield self.prepos[repo_name]

    def __contains__(self, repo_name):
        return repo_name in self.prepos

    def config_string(self):
        bool_keys = (
            "strict_misc_digests",
            "sync_allow_hardlinks",
            "sync_openpgp_key_refresh",
            "sync_rcu",
            "volatile",
        )
        str_or_int_keys = (
            "auto_sync",
            "clone_depth",
            "format",
            "location",
            "main_repo",
            "priority",
            "sync_depth",
            "sync_openpgp_keyserver",
            "sync_openpgp_key_path",
            "sync_openpgp_key_refresh_retry_count",
            "sync_openpgp_key_refresh_retry_delay_exp_base",
            "sync_openpgp_key_refresh_retry_delay_max",
            "sync_openpgp_key_refresh_retry_delay_mult",
            "sync_openpgp_key_refresh_retry_overall_timeout",
            "sync_rcu_spare_snapshots",
            "sync_rcu_store_dir",
            "sync_rcu_ttl_days",
            "sync_type",
            "sync_umask",
            "sync_uri",
            "sync_user",
        )
        str_tuple_keys = (
            "aliases",
            "eclass_overrides",
            "force",
        )
        repo_config_tuple_keys = ("masters",)
        keys = bool_keys + str_or_int_keys + str_tuple_keys + repo_config_tuple_keys
        config_string = ""
        for repo_name, repo in sorted(
            self.prepos.items(), key=lambda x: (x[0] != "DEFAULT", x[0])
        ):
            if repo_name != repo.name:
                continue
            config_string += f"\n[{repo_name}]\n"
            for key in sorted(keys):
                if key == "main_repo" and repo_name != "DEFAULT":
                    continue
                if getattr(repo, key) is not None:
                    if key in bool_keys:
                        config_string += "{} = {}\n".format(
                            key.replace("_", "-"),
                            "true" if getattr(repo, key) else "false",
                        )
                    elif key in str_or_int_keys:
                        config_string += "{} = {}\n".format(
                            key.replace("_", "-"),
                            getattr(repo, key),
                        )
                    elif key in str_tuple_keys:
                        config_string += "{} = {}\n".format(
                            key.replace("_", "-"),
                            " ".join(getattr(repo, key)),
                        )
                    elif key in repo_config_tuple_keys:
                        config_string += "{} = {}\n".format(
                            key.replace("_", "-"),
                            " ".join(x.name for x in getattr(repo, key)),
                        )
            for o, v in repo.module_specific_options.items():
                config_string += f"{o} = {v}\n"
        return config_string.lstrip("\n")


def allow_profile_repo_deps(
    repo: typing.Union[RepoConfig, _profile_node],
) -> bool:
    if eapi_has_repo_deps(repo.eapi):
        return True

    if "profile-repo-deps" in repo.profile_formats:
        return True

    return False


def load_repository_config(settings, extra_files=None):
    repoconfigpaths = []
    if "PORTAGE_REPOSITORIES" in settings:
        repoconfigpaths.append(io.StringIO(settings["PORTAGE_REPOSITORIES"]))
    else:
        if portage._not_installed:
            repoconfigpaths.append(os.path.join(PORTAGE_BASE_PATH, "cnf", "repos.conf"))
        else:
            repoconfigpaths.append(
                os.path.join(settings.global_config_path, "repos.conf")
            )
        repoconfigpaths.append(
            os.path.join(settings["PORTAGE_CONFIGROOT"], USER_CONFIG_PATH, "repos.conf")
        )
    if extra_files:
        repoconfigpaths.extend(extra_files)
    return RepoConfigLoader(repoconfigpaths, settings)


def _get_repo_name(repo_location, cached=None):
    if cached is not None:
        return cached
    name, missing = RepoConfig._read_repo_name(repo_location)
    if missing:
        return None
    return name


def parse_layout_conf(repo_location, repo_name=None):
    eapi = read_corresponding_eapi_file(os.path.join(repo_location, REPO_NAME_LOC))

    layout_filename = os.path.join(repo_location, "metadata", "layout.conf")
    layout_file = KeyValuePairFileLoader(layout_filename, None, None)
    layout_data, layout_errors = layout_file.load()

    data = {}

    # None indicates absence of a masters setting, which later code uses
    # to trigger a backward compatibility fallback that sets an implicit
    # master. In order to avoid this fallback behavior, layout.conf can
    # explicitly set masters to an empty value, which will result in an
    # empty tuple here instead of None.
    masters = layout_data.get("masters")
    if masters is not None:
        masters = tuple(masters.split())
    data["masters"] = masters
    data["aliases"] = tuple(layout_data.get("aliases", "").split())

    data["eapis-banned"] = tuple(layout_data.get("eapis-banned", "").split())
    data["eapis-deprecated"] = tuple(layout_data.get("eapis-deprecated", "").split())

    properties_allowed = layout_data.get("properties-allowed")
    if properties_allowed is not None:
        properties_allowed = tuple(properties_allowed.split())
    data["properties-allowed"] = properties_allowed

    restrict_allowed = layout_data.get("restrict-allowed")
    if restrict_allowed is not None:
        restrict_allowed = tuple(restrict_allowed.split())
    data["restrict-allowed"] = restrict_allowed

    data["sign-commit"] = layout_data.get("sign-commits", "false").lower() == "true"

    data["sign-manifest"] = layout_data.get("sign-manifests", "true").lower() == "true"

    data["thin-manifest"] = layout_data.get("thin-manifests", "false").lower() == "true"

    data["repo-name"] = _gen_valid_repo(layout_data.get("repo-name", ""))

    manifest_policy = layout_data.get("use-manifests", "strict").lower()
    data["allow-missing-manifest"] = manifest_policy != "strict"
    data["create-manifest"] = manifest_policy != "false"
    data["disable-manifest"] = manifest_policy == "false"

    # for compatibility w/ PMS, fallback to pms; but also check if the
    # cache exists or not.
    cache_formats = layout_data.get("cache-formats", "").lower().split()
    if not cache_formats:
        # Auto-detect cache formats, and prefer md5-cache if available.
        # This behavior was deployed in portage-2.1.11.14, so that the
        # default egencache format could eventually be changed to md5-dict
        # in portage-2.1.11.32. WARNING: Versions prior to portage-2.1.11.14
        # will NOT recognize md5-dict format unless it is explicitly
        # listed in layout.conf.
        cache_formats = []
        if os.path.isdir(os.path.join(repo_location, "metadata", "md5-cache")):
            cache_formats.append("md5-dict")
        if os.path.isdir(os.path.join(repo_location, "metadata", "cache")):
            cache_formats.append("pms")
    data["cache-formats"] = tuple(cache_formats)

    manifest_hashes = layout_data.get("manifest-hashes")
    manifest_required_hashes = layout_data.get("manifest-required-hashes")

    if manifest_required_hashes is not None and manifest_hashes is None:
        repo_name = _get_repo_name(repo_location, cached=repo_name)
        warnings.warn(
            (
                _(
                    "Repository named '%(repo_name)s' specifies "
                    "'manifest-required-hashes' setting without corresponding "
                    "'manifest-hashes'. Portage will default it to match "
                    "the required set but please add the missing entry "
                    "to: %(layout_filename)s"
                )
                % {
                    "repo_name": repo_name or "unspecified",
                    "layout_filename": layout_filename,
                }
            ),
            SyntaxWarning,
        )
        manifest_hashes = manifest_required_hashes

    if manifest_hashes is not None:
        # require all the hashes unless specified otherwise
        if manifest_required_hashes is None:
            manifest_required_hashes = manifest_hashes

        manifest_required_hashes = frozenset(manifest_required_hashes.upper().split())
        manifest_hashes = frozenset(manifest_hashes.upper().split())
        missing_required_hashes = manifest_required_hashes.difference(manifest_hashes)
        if missing_required_hashes:
            repo_name = _get_repo_name(repo_location, cached=repo_name)
            warnings.warn(
                (
                    _(
                        "Repository named '%(repo_name)s' has a "
                        "'manifest-hashes' setting that does not contain "
                        "the '%(hash)s' hashes which are listed in "
                        "'manifest-required-hashes'. Please fix that file "
                        "if you want to generate valid manifests for this "
                        "repository: %(layout_filename)s"
                    )
                    % {
                        "repo_name": repo_name or "unspecified",
                        "hash": " ".join(missing_required_hashes),
                        "layout_filename": layout_filename,
                    }
                ),
                SyntaxWarning,
            )
        unsupported_hashes = manifest_hashes.difference(get_valid_checksum_keys())
        if unsupported_hashes:
            repo_name = _get_repo_name(repo_location, cached=repo_name)
            warnings.warn(
                (
                    _(
                        "Repository named '%(repo_name)s' has a "
                        "'manifest-hashes' setting that contains one "
                        "or more hash types '%(hashes)s' which are not supported by "
                        "this portage version. You will have to upgrade "
                        "portage if you want to generate valid manifests for "
                        "this repository: %(layout_filename)s"
                    )
                    % {
                        "repo_name": repo_name or "unspecified",
                        "hashes": " ".join(sorted(unsupported_hashes)),
                        "layout_filename": layout_filename,
                    }
                ),
                DeprecationWarning,
            )

    data["manifest-hashes"] = manifest_hashes
    data["manifest-required-hashes"] = manifest_required_hashes

    data["update-changelog"] = (
        layout_data.get("update-changelog", "false").lower() == "true"
    )

    raw_formats = layout_data.get("profile-formats")
    if raw_formats is None:
        if eapi_allows_directories_on_profile_level_and_repository_level(eapi):
            raw_formats = ("portage-1",)
        else:
            raw_formats = ("portage-1-compat",)
    else:
        raw_formats = set(raw_formats.split())
        unknown = raw_formats.difference(_valid_profile_formats)
        if unknown:
            repo_name = _get_repo_name(repo_location, cached=repo_name)
            warnings.warn(
                (
                    _(
                        "Repository named '%(repo_name)s' has unsupported "
                        "profiles in use ('profile-formats = %(unknown_fmts)s' setting in "
                        "'%(layout_filename)s; please upgrade portage."
                    )
                    % dict(
                        repo_name=repo_name or "unspecified",
                        layout_filename=layout_filename,
                        unknown_fmts=" ".join(unknown),
                    )
                ),
                DeprecationWarning,
            )
        raw_formats = tuple(raw_formats.intersection(_valid_profile_formats))
    data["profile-formats"] = raw_formats

    try:
        eapi = layout_data["profile_eapi_when_unspecified"]
    except KeyError:
        pass
    else:
        if "profile-default-eapi" not in raw_formats:
            warnings.warn(
                (
                    _(
                        "Repository named '%(repo_name)s' has "
                        "profile_eapi_when_unspecified setting in "
                        "'%(layout_filename)s', but 'profile-default-eapi' is "
                        "not listed in the profile-formats field. Please "
                        "report this issue to the repository maintainer."
                    )
                    % dict(
                        repo_name=repo_name or "unspecified",
                        layout_filename=layout_filename,
                    )
                ),
                SyntaxWarning,
            )
        elif not portage.eapi_is_supported(eapi):
            warnings.warn(
                (
                    _(
                        "Repository named '%(repo_name)s' has "
                        "unsupported EAPI '%(eapi)s' setting in "
                        "'%(layout_filename)s'; please upgrade portage."
                    )
                    % dict(
                        repo_name=repo_name or "unspecified",
                        eapi=eapi,
                        layout_filename=layout_filename,
                    )
                ),
                SyntaxWarning,
            )
        else:
            data["profile_eapi_when_unspecified"] = eapi

    return data, layout_errors



# Copyright 2005-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import logging
import re
import subprocess
import datetime

import portage
from portage import os
from portage.util import writemsg_level, shlex_split
from portage.util.futures import asyncio
from portage.output import create_color_func, EOutput
from portage.const import TIMESTAMP_FORMAT

good = create_color_func("GOOD")
bad = create_color_func("BAD")
warn = create_color_func("WARN")
from portage.sync.syncbase import NewBase

try:
    from gemato.exceptions import GematoException
    import gemato.openpgp
except ImportError:
    gemato = None


class GitSync(NewBase):
    """Git sync class"""

    short_desc = "Perform sync operations on git based repositories"

    @staticmethod
    def name():
        return "GitSync"

    def __init__(self):
        NewBase.__init__(self, "git", portage.const.GIT_PACKAGE_ATOM)

    def exists(self, **kwargs) -> bool:
        """Tests whether the repo actually exists"""
        return os.path.exists(os.path.join(self.repo.location, ".git"))

    def new(self, **kwargs) -> tuple[int, bool]:
        """Do the initial clone of the repository"""
        if kwargs:
            self._kwargs(kwargs)
        if not self.has_bin:
            return (1, False)
        try:
            if not os.path.exists(self.repo.location):
                os.makedirs(self.repo.location)
                self.logger(
                    self.xterm_titles, f"Created new directory {self.repo.location}"
                )
        except OSError:
            return (1, False)

        sync_uri = self.repo.sync_uri
        if sync_uri.startswith("file://"):
            sync_uri = sync_uri[7:]

        git_cmd_opts = ""
        if self.repo.module_specific_options.get("sync-git-env"):
            shlexed_env = shlex_split(self.repo.module_specific_options["sync-git-env"])
            env = {
                k: v
                for k, _, v in (assignment.partition("=") for assignment in shlexed_env)
                if k
            }
            self.spawn_kwargs["env"].update(env)

        if self.repo.module_specific_options.get("sync-git-clone-env"):
            shlexed_env = shlex_split(
                self.repo.module_specific_options["sync-git-clone-env"]
            )
            clone_env = {
                k: v
                for k, _, v in (assignment.partition("=") for assignment in shlexed_env)
                if k
            }
            self.spawn_kwargs["env"].update(clone_env)

        if self.settings.get("PORTAGE_QUIET") == "1":
            git_cmd_opts += " --quiet"
        if self.repo.clone_depth is not None:
            if self.repo.clone_depth != 0:
                git_cmd_opts += " --depth %d" % self.repo.clone_depth
        else:
            # default
            git_cmd_opts += " --depth 1"

        if self.repo.module_specific_options.get("sync-git-clone-extra-opts"):
            git_cmd_opts += (
                f" {self.repo.module_specific_options['sync-git-clone-extra-opts']}"
            )
        git_cmd = "{} clone{} {} .".format(
            self.bin_command,
            git_cmd_opts,
            portage._shell_quote(sync_uri),
        )
        writemsg_level(git_cmd + "\n")

        exitcode = portage.process.spawn_bash(
            f"cd {portage._shell_quote(self.repo.location)} ; exec {git_cmd}",
            **self.spawn_kwargs,
        )
        if exitcode != os.EX_OK:
            msg = f"!!! git clone error in {self.repo.location}"
            self.logger(self.xterm_titles, msg)
            writemsg_level(msg + "\n", level=logging.ERROR, noiselevel=-1)
            return (exitcode, False)

        self.add_safe_directory()

        if not self.verify_head():
            return (1, False)

        return (os.EX_OK, True)

    def _gen_ceiling_string(self, path: str) -> str:
        """
        Iteratively generate a colon delimited string of all of the
        given path's parents, for use with GIT_CEILING_DIRECTORIES
        """
        directories = []

        while True:
            if path == "/":
                break
            path = os.path.dirname(path)
            directories.append(path)

        return ":".join(directories)

    def update(self) -> tuple[int, bool]:
        """Update existing git repository, and ignore the syncuri. We are
        going to trust the user and assume that the user is in the branch
        that he/she wants updated. We'll let the user manage branches with
        git directly.
        """
        if not self.has_bin:
            return (1, False)

        opts = self.options.get("emerge_config").opts

        git_cmd_opts = ""
        quiet = self.settings.get("PORTAGE_QUIET") == "1"
        verbose = "--verbose" in opts

        # We don't want to operate with a .git outside of the given
        # repo in any circumstances.
        self.spawn_kwargs["env"].update(
            {"GIT_CEILING_DIRECTORIES": self._gen_ceiling_string(self.repo.location)}
        )

        self.add_safe_directory()

        if self.repo.module_specific_options.get("sync-git-env"):
            shlexed_env = shlex_split(self.repo.module_specific_options["sync-git-env"])
            env = {
                k: v
                for k, _, v in (assignment.partition("=") for assignment in shlexed_env)
                if k
            }
            self.spawn_kwargs["env"].update(env)

        if self.repo.module_specific_options.get("sync-git-pull-env"):
            shlexed_env = shlex_split(
                self.repo.module_specific_options["sync-git-pull-env"]
            )
            pull_env = {
                k: v
                for k, _, v in (assignment.partition("=") for assignment in shlexed_env)
                if k
            }
            self.spawn_kwargs["env"].update(pull_env)

        if quiet:
            git_cmd_opts += " --quiet"
        elif verbose:
            git_cmd_opts += " --verbose"

        # The logic here is a bit delicate. We need to balance two things:
        # 1. Having a robust sync mechanism which works unattended.
        # 2. Allowing users to have the flexibility they might expect when using
        # a git repository in repos.conf for syncing.
        #
        # For sync-type=git repositories, we've seen a problem in the wild
        # where shallow clones end up "breaking themselves" especially when
        # the origin is behind a CDN. 'git pull' might return state X,
        # but on a subsequent pull, return state X-1. git will then (sometimes)
        # leave orphaned untracked files in the repository. On a subsequent pull,
        # when state >= X is returned where those files exist in the origin,
        # git then refuses to write over them and aborts to avoid clobbering
        # local work.
        #
        # To mitigate this, Portage will aggressively clobber any changes
        # in the local directory, as its priority is to keep syncing working,
        # by running 'git clean' and 'git reset --hard'.
        #
        # Portage performs this clobbering if:
        # 1. sync-type=git
        # 2.
        #   - volatile=no (explicitly set to no), OR
        #   - volatile is unset AND the repository owner is either root or portage
        # 3. Portage is syncing the repository (rather than e.g. auto-sync=no
        # and never running 'emaint sync -r foo')
        #
        # Portage will not clobber if:
        # 1. volatile=yes (explicitly set in the config), OR
        # 2. volatile is unset and the repository owner is neither root nor
        #    portage.
        #
        # 'volatile' refers to whether the repository is volatile and may
        # only be safely changed by Portage itself, i.e. whether Portage
        # should expect the user to change it or not.
        #
        # - volatile=yes:
        # The repository is volatile and may be changed at any time by the user.
        # Portage will not perform destructive operations on the repository.
        # - volatile=no
        # The repository is not volatile. Only Portage may modify the
        # repository. User changes may be lost.
        # Portage may perform destructive operations on the repository
        # to keep sync working.
        #
        # References:
        # bug #887025
        # bug #824782
        # https://archives.gentoo.org/gentoo-dev/message/f58a97027252458ad0a44090a2602897

        # Default: Perform shallow updates (but only if the target is
        # already a shallow repository).
        sync_depth = 1
        if self.repo.sync_depth is not None:
            sync_depth = self.repo.sync_depth
        else:
            if self.repo.volatile:
                # If sync-depth is not explicitly set by the user,
                # then check if the target repository is already a
                # shallow one. And do not perform a shallow update if
                # the target repository is not shallow.
                is_shallow_cmd = ["git", "rev-parse", "--is-shallow-repository"]
                is_shallow_res = portage._unicode_decode(
                    subprocess.check_output(
                        is_shallow_cmd,
                        cwd=portage._unicode_encode(self.repo.location),
                    )
                ).rstrip("\n")
                if is_shallow_res == "false":
                    sync_depth = 0
            else:
                # If the repository is marked as non-volatile, we assume
                # it's fine to Portage to do what it wishes to it.
                sync_depth = 1

        shallow = False
        if sync_depth > 0:
            git_cmd_opts += f" --depth {sync_depth}"
            shallow = True

        if self.repo.module_specific_options.get("sync-git-pull-extra-opts"):
            git_cmd_opts += (
                f" {self.repo.module_specific_options['sync-git-pull-extra-opts']}"
            )

        try:
            remote_branch = portage._unicode_decode(
                subprocess.check_output(
                    [
                        self.bin_command,
                        "rev-parse",
                        "--abbrev-ref",
                        "--symbolic-full-name",
                        "@{upstream}",
                    ],
                    cwd=portage._unicode_encode(self.repo.location),
                )
            ).rstrip("\n")
        except subprocess.CalledProcessError as e:
            msg = f"!!! git rev-parse error in {self.repo.location}"
            self.logger(self.xterm_titles, msg)
            writemsg_level(msg + "\n", level=logging.ERROR, noiselevel=-1)
            return (e.returncode, False)

        if shallow:
            # For shallow fetch, unreachable objects may need to be pruned
            # manually, in order to prevent automatic git gc calls from
            # eventually failing (see bug 599008).
            gc_cmd = ["git", "-c", "gc.autodetach=false", "gc", "--auto"]
            if quiet:
                gc_cmd.append("--quiet")
            exitcode = portage.process.spawn(
                gc_cmd,
                cwd=portage._unicode_encode(self.repo.location),
                **self.spawn_kwargs,
            )
            if exitcode != os.EX_OK:
                msg = f"!!! git gc error in {self.repo.location}"
                self.logger(self.xterm_titles, msg)
                writemsg_level(msg + "\n", level=logging.ERROR, noiselevel=-1)
                return (exitcode, False)

        git_remote = remote_branch.partition("/")[0]

        if not self.repo.volatile:
            git_get_remote_url_cmd = ["git", "ls-remote", "--get-url", git_remote]
            git_remote_url = portage._unicode_decode(
                subprocess.check_output(
                    git_get_remote_url_cmd,
                    cwd=portage._unicode_encode(self.repo.location),
                )
            ).strip()
            if git_remote_url != self.repo.sync_uri:
                git_set_remote_url_cmd = [
                    "git",
                    "remote",
                    "set-url",
                    git_remote,
                    self.repo.sync_uri,
                ]
                exitcode = portage.process.spawn(
                    git_set_remote_url_cmd,
                    cwd=portage._unicode_encode(self.repo.location),
                    **self.spawn_kwargs,
                )
                if exitcode != os.EX_OK:
                    msg = f"!!! could not update git remote {git_remote}'s url to {self.repo.sync_uri}"
                    self.logger(self.xterm_titles, msg)
                    writemsg_level(msg + "\n", level=logging.ERROR, noiselevel=-1)
                    return (exitcode, False)
                elif not quiet:
                    writemsg_level(" ".join(git_set_remote_url_cmd) + "\n")

        git_cmd = f"{self.bin_command} fetch {git_remote}{git_cmd_opts}"

        if not quiet:
            writemsg_level(git_cmd + "\n")

        rev_cmd = [self.bin_command, "rev-list", "--max-count=1", "HEAD"]
        previous_rev = subprocess.check_output(
            rev_cmd, cwd=portage._unicode_encode(self.repo.location)
        )

        exitcode = portage.process.spawn_bash(
            f"cd {portage._shell_quote(self.repo.location)} ; exec {git_cmd}",
            **self.spawn_kwargs,
        )

        if exitcode != os.EX_OK:
            msg = f"!!! git fetch error in {self.repo.location}"
            self.logger(self.xterm_titles, msg)
            writemsg_level(msg + "\n", level=logging.ERROR, noiselevel=-1)
            return (exitcode, False)

        if not self.verify_head(revision=f"refs/remotes/{remote_branch}"):
            return (1, False)

        if not self.repo.volatile:
            # Clean up the repo before trying to sync to upstream.
            # - Only done for volatile=false repositories to avoid losing
            # data.
            # - This is needed to avoid orphaned files preventing further syncs
            # on shallow clones.
            clean_cmd = [self.bin_command, "clean", "--force", "-d", "-x"]

            if quiet:
                clean_cmd.append("--quiet")

            exitcode = portage.process.spawn(
                clean_cmd,
                cwd=portage._unicode_encode(self.repo.location),
                **self.spawn_kwargs,
            )

            if exitcode != os.EX_OK:
                msg = f"!!! git clean error in {self.repo.location}"
                self.logger(self.xterm_titles, msg)
                writemsg_level(msg + "\n", level=logging.ERROR, noiselevel=-1)
                return (exitcode, False)

        # `git diff --quiet` returns 0 on a clean tree and 1 otherwise
        is_clean = (
            portage.process.spawn(
                f"{self.bin_command} diff --quiet",
                cwd=portage._unicode_encode(self.repo.location),
                **self.spawn_kwargs,
            )
            == 0
        )

        if not is_clean and not self.repo.volatile:
            # If the repo isn't clean, clobber any changes for parity
            # with rsync. Only do this for non-volatile repositories.
            merge_cmd = [self.bin_command, "reset", "--hard"]
        elif shallow:
            # Since the default merge strategy typically fails when
            # the depth is not unlimited, `git reset --merge`.
            merge_cmd = [self.bin_command, "reset", "--merge"]
        else:
            merge_cmd = [self.bin_command, "merge"]

        merge_cmd.append(f"refs/remotes/{remote_branch}")
        if quiet:
            merge_cmd.append("--quiet")

        if not quiet:
            writemsg_level(" ".join(merge_cmd) + "\n")

        exitcode = portage.process.spawn(
            merge_cmd,
            cwd=portage._unicode_encode(self.repo.location),
            **self.spawn_kwargs,
        )

        if exitcode != os.EX_OK:
            if not self.repo.volatile:
                # HACK - sometimes merging results in a tree diverged from
                # upstream, so try to hack around it
                # https://stackoverflow.com/questions/41075972/how-to-update-a-git-shallow-clone/41081908#41081908
                exitcode = portage.process.spawn(
                    f"{self.bin_command} reset --hard refs/remotes/{remote_branch}",
                    cwd=portage._unicode_encode(self.repo.location),
                    **self.spawn_kwargs,
                )

            if exitcode != os.EX_OK:
                msg = f"!!! git merge error in {self.repo.location}"
                self.logger(self.xterm_titles, msg)
                writemsg_level(msg + "\n", level=logging.ERROR, noiselevel=-1)
                return (exitcode, False)

        current_rev = subprocess.check_output(
            rev_cmd, cwd=portage._unicode_encode(self.repo.location)
        )

        return (os.EX_OK, current_rev != previous_rev)

    def verify_head(self, revision="-1") -> bool:
        max_age_days = self.repo.module_specific_options.get(
            "sync-git-verify-max-age-days", ""
        )
        if max_age_days:
            try:
                max_age_days = int(max_age_days)
                if max_age_days <= 0:
                    raise ValueError(max_age_days)
            except ValueError:
                writemsg_level(
                    f"!!! sync-git-max-age-days must be a positive non-zero integer: {max_age_days}\n",
                    level=logging.ERROR,
                    noiselevel=-1,
                )
                return False
            show_timestamp_chk_file_cmd = [
                self.bin_command,
                "show",
                f"{revision}:metadata/timestamp.chk",
            ]
            try:
                timestamp_chk = portage._unicode_decode(
                    subprocess.check_output(
                        show_timestamp_chk_file_cmd,
                        cwd=portage._unicode_encode(self.repo.location),
                    )
                ).strip()
            except subprocess.CalledProcessError as e:
                writemsg_level(
                    f"!!! {show_timestamp_chk_file_cmd} failed with {e.returncode}",
                    level=logging.ERROR,
                    noiselevel=-1,
                )
                return False
            timestamp = datetime.datetime.strptime(timestamp_chk, TIMESTAMP_FORMAT)
            max_timestamp_age = datetime.datetime.now() - datetime.timedelta(
                days=max_age_days
            )
            if timestamp < max_timestamp_age:
                writemsg_level(
                    f"!!! timestamp (from timestamp.chk) {timestamp} is older than max age {max_timestamp_age}\n",
                    level=logging.ERROR,
                    noiselevel=-1,
                )
                return False

        if self.repo.module_specific_options.get(
            "sync-git-verify-commit-signature", "false"
        ).lower() not in ("true", "yes"):
            return True

        if self.repo.sync_openpgp_key_path is not None and gemato is None:
            writemsg_level(
                "!!! Verifying against specified key requires gemato-14.5+ installed\n",
                level=logging.ERROR,
                noiselevel=-1,
            )
            return False

        opts = self.options.get("emerge_config").opts
        debug = "--debug" in opts
        quiet = self.settings.get("PORTAGE_QUIET") == "1"
        verbose = "--verbose" in opts

        openpgp_env = self._get_openpgp_env(self.repo.sync_openpgp_key_path, debug)

        if debug:
            old_level = logging.getLogger().getEffectiveLevel()
            logging.getLogger().setLevel(logging.DEBUG)
            logging.getLogger("gemato").setLevel(logging.DEBUG)

        try:
            out = EOutput()
            env = None
            if openpgp_env is not None and self.repo.sync_openpgp_key_path is not None:
                try:
                    out.einfo(f"Using keys from {self.repo.sync_openpgp_key_path}")
                    with open(self.repo.sync_openpgp_key_path, "rb") as f:
                        openpgp_env.import_key(f)
                    self._refresh_keys(openpgp_env)
                except (GematoException, asyncio.TimeoutError) as e:
                    writemsg_level(
                        f"!!! Verification impossible due to keyring problem:\n{e}\n",
                        level=logging.ERROR,
                        noiselevel=-1,
                    )
                    return False

                env = os.environ.copy()
                env["GNUPGHOME"] = openpgp_env.home

            rev_cmd = [
                self.bin_command,
                "-c",
                "log.showsignature=0",
                "log",
                "-n1",
                "--pretty=format:%G?%n%GF",
                revision,
            ]
            try:
                lines = portage._unicode_decode(
                    subprocess.check_output(
                        rev_cmd,
                        cwd=portage._unicode_encode(self.repo.location),
                        env=env,
                    )
                ).splitlines()
            except subprocess.CalledProcessError:
                return False

            status = lines[0].strip()
            if len(lines) > 1:
                signing_key = lines[1].strip()

            if status == "G":  # good signature is good
                if not quiet:
                    message = "Trusted signature found on top commit"
                    if verbose:
                        message += (
                            f" (git revision: {revision}, signing key: {signing_key})"
                        )
                    out.einfo(message)
                return True
            if status == "U":  # untrusted
                out.ewarn(
                    f"Top commit signature is valid but not trusted (git revision: {revision}, signing key: {signing_key})"
                )
                return True
            if status == "B":
                expl = (
                    f"bad signature using key {signing_key} on git revision {revision}"
                )
            elif status == "X":
                expl = f"expired signature using key {signing_key} on git revision {revision}"
            elif status == "Y":
                expl = f"expired key using key {signing_key} on git revision {revision}"
            elif status == "R":
                expl = f"revoked key using key {signing_key} on git revision {revision}"
            elif status == "E":
                expl = "unable to verify signature (missing key?)"
            elif status == "N":
                expl = "no signature"
            else:
                expl = "unknown issue"
            out.eerror(f"No valid signature found: {expl}")

            if debug:
                writemsg_level(
                    f"!!! Got following output from gpg: {status}\n",
                    level=logging.DEBUG,
                    noiselevel=-1,
                )

            return False
        finally:
            if openpgp_env is not None:
                openpgp_env.close()
            if debug:
                logging.getLogger().setLevel(old_level)

    def retrieve_head(self, **kwargs) -> tuple[int, bool]:
        """Get information about the head commit"""
        if kwargs:
            self._kwargs(kwargs)
        if self.bin_command is None:
            # return quietly so that we don't pollute emerge --info output
            return (1, False)
        rev_cmd = [self.bin_command, "rev-list", "--max-count=1", "HEAD"]
        try:
            ret = (
                os.EX_OK,
                portage._unicode_decode(
                    subprocess.check_output(
                        rev_cmd, cwd=portage._unicode_encode(self.repo.location)
                    )
                ),
            )
        except subprocess.CalledProcessError:
            ret = (1, False)
        return ret

    def add_safe_directory(self) -> bool:
        # Add safe.directory to system gitconfig if not already configured.
        # Workaround for bug #838271 and bug #838223.
        location_escaped = re.escape(self.repo.location)
        result = subprocess.run(
            [
                self.bin_command,
                "config",
                "--get",
                "safe.directory",
                f"^{location_escaped}$",
            ],
            stdout=subprocess.DEVNULL,
        )
        if result.returncode == 1:
            result = subprocess.run(
                [
                    self.bin_command,
                    "config",
                    "--system",
                    "--add",
                    "safe.directory",
                    self.repo.location,
                ],
                stdout=subprocess.DEVNULL,
            )
        return result.returncode == 0



# Copyright 1999-2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.util.SlotObject import SlotObject


class Task(SlotObject):
    __slots__ = ("_hash_key", "_hash_value")

    def __eq__(self, other):
        try:
            return self._hash_key == other._hash_key
        except AttributeError:
            # depgraph._pkg() generates _hash_key
            # for lookups here, so handle that
            return self._hash_key == other

    def __ne__(self, other):
        try:
            return self._hash_key != other._hash_key
        except AttributeError:
            return True

    def __hash__(self):
        return self._hash_value

    def __len__(self):
        return len(self._hash_key)

    def __getitem__(self, key):
        return self._hash_key[key]

    def __iter__(self):
        return iter(self._hash_key)

    def __contains__(self, key):
        return key in self._hash_key

    def __str__(self):
        """
        Emulate tuple.__repr__, but don't show 'foo' as u'foo' for unicode
        strings.
        """
        strings = (f"'{x}'" for x in self._hash_key)
        return f"({', '.join(strings)})"

    def __repr__(self):
        if self._hash_key is None:
            # triggered by python-trace
            return SlotObject.__repr__(self)
        return "<{} ({})>".format(
            self.__class__.__name__,
            ", ".join(f"'{x}'" for x in self._hash_key),
        )



# Copyright 1999-2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import re
from portage.dep import isvalidatom


def insert_category_into_atom(atom, category):
    # Handle '*' character for "extended syntax" wildcard support.
    alphanum = re.search(r"[\*\w]", atom, re.UNICODE)
    if alphanum:
        ret = atom[: alphanum.start()] + f"{category}/" + atom[alphanum.start() :]
    else:
        ret = None
    return ret


def is_valid_package_atom(x, allow_repo=False, allow_build_id=True):
    if "/" not in x.split(":")[0]:
        x2 = insert_category_into_atom(x, "cat")
        if x2 is not None:
            x = x2
    return isvalidatom(
        x, allow_blockers=False, allow_repo=allow_repo, allow_build_id=allow_build_id
    )



# Copyright 2020-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import functools
import multiprocessing

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.util.futures import asyncio
from portage.util.futures.executor.fork import ForkExecutor


class AuxdbTestCase(TestCase):
    def test_anydbm(self):
        try:
            from portage.cache.anydbm import database
        except ImportError:
            self.skipTest("dbm import failed")
        self._test_mod(
            "portage.cache.anydbm.database", multiproc=False, picklable=False
        )

    def test_flat_hash_md5(self):
        self._test_mod("portage.cache.flat_hash.md5_database")

    def test_volatile(self):
        self._test_mod("portage.cache.volatile.database", multiproc=False)

    def test_sqite(self):
        try:
            import sqlite3
        except ImportError:
            self.skipTest("sqlite3 import failed")
        self._test_mod("portage.cache.sqlite.database", picklable=True)

    def _test_mod(self, auxdbmodule, multiproc=True, picklable=True):
        ebuilds = {
            "cat/A-1": {
                "EAPI": "7",
                "MISC_CONTENT": "inherit foo",
            },
            "cat/B-1": {
                "EAPI": "7",
                "MISC_CONTENT": "inherit foo",
            },
        }

        ebuild_inherited = frozenset(["bar", "foo"])
        eclass_defined_phases = "prepare"
        eclass_depend = "bar/foo"

        eclasses = {
            "foo": ("inherit bar",),
            "bar": (
                "EXPORT_FUNCTIONS src_prepare",
                f'DEPEND="{eclass_depend}"',
                "bar_src_prepare() { default; }",
            ),
        }

        playground = ResolverPlayground(
            ebuilds=ebuilds,
            eclasses=eclasses,
            user_config={"modules": (f"portdbapi.auxdbmodule = {auxdbmodule}",)},
        )

        portdb = playground.trees[playground.eroot]["porttree"].dbapi
        metadata_keys = ["DEFINED_PHASES", "DEPEND", "EAPI", "INHERITED"]

        test_func = functools.partial(
            self._run_test_mod_async, ebuilds, metadata_keys, portdb
        )

        results = test_func()

        self._compare_results(
            ebuilds, eclass_defined_phases, eclass_depend, ebuild_inherited, results
        )

        loop = asyncio._wrap_loop()
        picklable_or_fork = picklable or multiprocessing.get_start_method == "fork"
        if picklable_or_fork:
            results = loop.run_until_complete(
                loop.run_in_executor(ForkExecutor(), test_func)
            )

            self._compare_results(
                ebuilds, eclass_defined_phases, eclass_depend, ebuild_inherited, results
            )

        auxdb = portdb.auxdb[portdb.getRepositoryPath("test_repo")]
        cpv = next(iter(ebuilds))

        modify_auxdb = functools.partial(self._modify_auxdb, auxdb, cpv)

        if multiproc and picklable_or_fork:
            loop.run_until_complete(loop.run_in_executor(ForkExecutor(), modify_auxdb))
        else:
            modify_auxdb()

        self.assertEqual(auxdb[cpv]["RESTRICT"], "test")

    def _compare_results(
        self, ebuilds, eclass_defined_phases, eclass_depend, ebuild_inherited, results
    ):
        for cpv, metadata in ebuilds.items():
            self.assertEqual(results[cpv]["DEFINED_PHASES"], eclass_defined_phases)
            self.assertEqual(results[cpv]["DEPEND"], eclass_depend)
            self.assertEqual(results[cpv]["EAPI"], metadata["EAPI"])
            self.assertEqual(
                frozenset(results[cpv]["INHERITED"].split()), ebuild_inherited
            )

    @staticmethod
    def _run_test_mod_async(ebuilds, metadata_keys, portdb):
        loop = asyncio._wrap_loop()
        return loop.run_until_complete(
            AuxdbTestCase._test_mod_async(
                ebuilds,
                metadata_keys,
                portdb,
            )
        )

    @staticmethod
    async def _test_mod_async(ebuilds, metadata_keys, portdb):
        results = {}
        for cpv, metadata in ebuilds.items():
            results[cpv] = dict(
                zip(metadata_keys, await portdb.async_aux_get(cpv, metadata_keys))
            )

        return results

    @staticmethod
    def _modify_auxdb(auxdb, cpv):
        metadata = auxdb[cpv]
        metadata["RESTRICT"] = "test"
        try:
            del metadata["_eclasses_"]
        except KeyError:
            pass
        auxdb[cpv] = metadata



# Copyright 2010 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class MergelistOutputTestCase(TestCase):
    def testMergelistOutput(self):
        """
        This test doesn't check if the output is correct, but makes sure
        that we don't backtrace somewhere in the output code.
        """
        ebuilds = {
            "dev-libs/A-1": {
                "DEPEND": "dev-libs/B dev-libs/C",
                "IUSE": "+foo",
                "EAPI": 1,
            },
            "dev-libs/B-1": {"DEPEND": "dev-libs/D", "IUSE": "foo +bar", "EAPI": 1},
            "dev-libs/C-1": {"DEPEND": "dev-libs/E", "IUSE": "foo bar"},
            "dev-libs/D-1": {"IUSE": ""},
            "dev-libs/E-1": {},
            # reinstall for flags
            "dev-libs/Z-1": {"IUSE": "+foo", "EAPI": 1},
            "dev-libs/Y-1": {"IUSE": "foo", "EAPI": 1},
            "dev-libs/X-1": {},
            "dev-libs/W-1": {"IUSE": "+foo", "EAPI": 1},
        }

        installed = {
            "dev-libs/Z-1": {"USE": "", "IUSE": "foo"},
            "dev-libs/Y-1": {"USE": "foo", "IUSE": "+foo", "EAPI": 1},
            "dev-libs/X-1": {"USE": "foo", "IUSE": "+foo", "EAPI": 1},
            "dev-libs/W-1": {},
        }

        option_cobos = (
            (),
            ("verbose",),
            ("tree",),
            (
                "tree",
                "unordered-display",
            ),
            ("verbose",),
            (
                "verbose",
                "tree",
            ),
            (
                "verbose",
                "tree",
                "unordered-display",
            ),
        )

        test_cases = []
        for options in option_cobos:
            testcase_opts = {}
            for opt in options:
                testcase_opts["--" + opt] = True

            test_cases.append(
                ResolverPlaygroundTestCase(
                    ["dev-libs/A"],
                    options=testcase_opts,
                    success=True,
                    ignore_mergelist_order=True,
                    mergelist=[
                        "dev-libs/D-1",
                        "dev-libs/E-1",
                        "dev-libs/C-1",
                        "dev-libs/B-1",
                        "dev-libs/A-1",
                    ],
                )
            )

            test_cases.append(
                ResolverPlaygroundTestCase(
                    ["dev-libs/Z"],
                    options=testcase_opts,
                    success=True,
                    mergelist=["dev-libs/Z-1"],
                )
            )

            test_cases.append(
                ResolverPlaygroundTestCase(
                    ["dev-libs/Y"],
                    options=testcase_opts,
                    success=True,
                    mergelist=["dev-libs/Y-1"],
                )
            )

            test_cases.append(
                ResolverPlaygroundTestCase(
                    ["dev-libs/X"],
                    options=testcase_opts,
                    success=True,
                    mergelist=["dev-libs/X-1"],
                )
            )

            test_cases.append(
                ResolverPlaygroundTestCase(
                    ["dev-libs/W"],
                    options=testcase_opts,
                    success=True,
                    mergelist=["dev-libs/W-1"],
                )
            )

        playground = ResolverPlayground(ebuilds=ebuilds, installed=installed)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2001-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import tarfile
import io
import threading
import subprocess
import errno
import pwd
import grp
import stat
import sys
import tempfile
from copy import copy
from datetime import datetime

import portage
from portage import checksum
from portage import os
from portage import shutil
from portage import normalize_path
from portage import _encodings
from portage import _unicode_decode
from portage import _unicode_encode
from portage.binpkg import get_binpkg_format
from portage.exception import (
    FileNotFound,
    InvalidBinaryPackageFormat,
    InvalidCompressionMethod,
    CompressorNotFound,
    CompressorOperationFailed,
    CommandNotFound,
    GPGException,
    DigestException,
    MissingSignature,
    InvalidSignature,
    SignedPackage,
)
from portage.output import colorize, EOutput
from portage.util._urlopen import urlopen
from portage.util import writemsg
from portage.util import shlex_split, varexpand
from portage.util.compression_probe import _compressors
from portage.util.cpuinfo import makeopts_to_job_count
from portage.process import find_binary
from portage.const import MANIFEST2_HASH_DEFAULTS, HASHING_BLOCKSIZE


class tar_stream_writer:
    """
    One-pass helper function that return a file-like object
    for create a file inside of a tar container.

    This helper allowed streaming add a new file to tar
    without prior knows the file size.

    With optional call and pipe data through external program,
    the helper can transparently save compressed data.

    With optional checksum helper, this helper can create
    corresponding checksum and GPG signature.

    Example:

    writer = tar_stream_writer(
            file_tarinfo,            # the file tarinfo that need to be added
            container,               # the outer container tarfile object
            tarfile.USTAR_FORMAT,    # the outer container format
            ["gzip"],                # compression command
            checksum_helper          # checksum helper
    )

    writer.write(data)
    writer.close()
    """

    def __init__(
        self,
        tarinfo,
        container,
        tar_format,
        cmd=None,
        checksum_helper=None,
        uid=None,
        gid=None,
    ):
        """
        tarinfo          # the file tarinfo that need to be added
        container        # the outer container tarfile object
        tar_format       # the outer container format for create the tar header
        cmd              # subprocess.Popen format compression command
        checksum_helper  # checksum helper
        uid              # drop root user to uid
        gid              # drop root group to gid
        """
        self.checksum_helper = checksum_helper
        self.cmd = cmd
        self.closed = False
        self.container = container
        self.killed = False
        self.tar_format = tar_format
        self.tarinfo = tarinfo
        self.uid = uid
        self.gid = gid

        # Record container end position
        self.container.fileobj.seek(0, io.SEEK_END)
        self.begin_position = self.container.fileobj.tell()
        self.end_position = 0
        self.file_size = 0

        # Write tar header without size
        tar_header = self.tarinfo.tobuf(
            self.tar_format, self.container.encoding, self.container.errors
        )
        self.header_size = len(tar_header)
        self.container.fileobj.write(tar_header)
        self.container.fileobj.flush()
        self.container.offset += self.header_size

        # Start external compressor if needed
        if cmd is None:
            self.proc = None
        else:
            self.proc = subprocess.Popen(
                cmd,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                user=self.uid,
                group=self.gid,
            )

            self.read_thread = threading.Thread(
                target=self._cmd_read_thread, name="tar_stream_cmd_read", daemon=True
            )
            self.read_thread.start()

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def kill(self):
        """
        kill external program if any error happened in python
        """
        if self.proc is not None:
            self.killed = True
            self.proc.kill()
            self.proc.stdin.close()
            self.close()

    def _cmd_read_thread(self):
        """
        Use thread to avoid block.
        Read stdout from external compressor, then write to the file
        in container, and to checksum helper if needed.
        """
        while True:
            try:
                buffer = self.proc.stdout.read(HASHING_BLOCKSIZE)
                if not buffer:
                    self.proc.stdout.close()
                    return
            except BrokenPipeError:
                self.proc.stdout.close()
                writemsg(colorize("BAD", f"GPKG subprocess failed: {self.cmd} \n"))
                if not self.killed:
                    # Do not raise error if killed by portage
                    raise CompressorOperationFailed("PIPE broken")
            try:
                self.container.fileobj.write(buffer)
            except OSError as err:
                self.error = True
                self.kill()
                raise CompressorOperationFailed(str(err))
            if self.checksum_helper:
                self.checksum_helper.update(buffer)

    def write(self, data):
        """
        Write data to tarfile or external compressor stdin
        """
        if self.closed:
            raise OSError("writer closed")

        if self.proc:
            # Write to external program
            try:
                self.proc.stdin.write(data)
            except BrokenPipeError:
                self.error = True
                writemsg(colorize("BAD", f"GPKG subprocess failed: {self.cmd} \n"))
                raise CompressorOperationFailed("PIPE broken")
        else:
            # Write to container
            self.container.fileobj.write(data)
            if self.checksum_helper:
                self.checksum_helper.update(data)

    def close(self):
        """
        Update the new file tar header when close
        """
        if self.closed:
            return

        # Wait compressor exit
        if self.proc is not None:
            self.proc.stdin.close()
            if self.proc.wait() != os.EX_OK:
                if not self.error:
                    raise CompressorOperationFailed("compression failed")
            if self.read_thread.is_alive():
                self.read_thread.join()

        # Get container end position and calculate file size
        self.container.fileobj.seek(0, io.SEEK_END)
        self.end_position = self.container.fileobj.tell()
        self.file_size = self.end_position - self.begin_position - self.header_size
        self.tarinfo.size = self.file_size

        # Tar block is 512, need padding \0
        _, remainder = divmod(self.file_size, 512)
        if remainder > 0:
            padding_size = 512 - remainder
            self.container.fileobj.write(b"\0" * padding_size)
            self.container.offset += padding_size
            self.container.fileobj.flush()

        # Update tar header
        tar_header = self.tarinfo.tobuf(
            self.tar_format, self.container.encoding, self.container.errors
        )
        self.container.fileobj.seek(self.begin_position)
        self.container.fileobj.write(tar_header)
        self.container.fileobj.seek(0, io.SEEK_END)
        self.container.fileobj.flush()
        self.container.offset = self.container.fileobj.tell()
        self.closed = True

        # Add tarinfo to tarfile
        self.container.members.append(self.tarinfo)

        if self.checksum_helper:
            self.checksum_helper.finish()

        self.closed = True


class tar_stream_reader:
    """
    helper function that return a file-like object
    for read a file inside of a tar container.

    This helper allowed transparently streaming read a compressed
    file in tar.

    With optional call and pipe compressed data through external
    program, and return the uncompressed data.

    reader = tar_stream_reader(
            fileobj,             # the fileobj from tarfile.extractfile(f)
            ["gzip", "-d"],      # decompression command
    )

    reader.read()
    reader.close()
    """

    def __init__(self, fileobj, cmd=None, uid=None, gid=None):
        """
        fileobj should be a file-like object that have read().
        cmd is optional external decompressor command.
        """
        self.closed = False
        self.cmd = cmd
        self.fileobj = fileobj
        self.killed = False
        self.uid = uid
        self.gid = gid

        if cmd is None:
            self.read_io = fileobj
            self.proc = None
        else:
            # Start external decompressor
            self.proc = subprocess.Popen(
                cmd,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                user=self.uid,
                group=self.gid,
            )
            self.read_io = self.proc.stdout
            # Start stdin block writing thread
            self.thread = threading.Thread(
                target=self._write_thread, name="tar_stream_stdin_writer", daemon=True
            )
            self.thread.start()

    def __del__(self):
        try:
            self.close()
        except CompressorOperationFailed:
            pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        try:
            self.close()
        except CompressorOperationFailed:
            pass

    def _write_thread(self):
        """
        writing thread to avoid full buffer blocking
        """
        try:
            while True:
                buffer = self.fileobj.read(HASHING_BLOCKSIZE)
                if buffer:
                    try:
                        self.proc.stdin.write(buffer)
                    except ValueError:
                        if self.killed:
                            return
                        else:
                            raise
                else:
                    self.proc.stdin.flush()
                    self.proc.stdin.close()
                    break
        except BrokenPipeError:
            if self.killed is False:
                writemsg(colorize("BAD", f"GPKG subprocess failed: {self.cmd} \n"))
                raise CompressorOperationFailed("PIPE broken")

    def kill(self):
        """
        kill external program if any error happened in python
        """
        if self.proc is not None:
            self.killed = True
            self.proc.kill()
            self.proc.stdin.close()
            self.close()

    def read(self, bufsize=-1):
        """
        return decompressor stdout data
        """
        if self.closed:
            raise OSError("writer closed")
        else:
            return self.read_io.read(bufsize)

    def close(self):
        """
        wait external program complete and do clean up
        """
        if self.closed:
            return

        self.closed = True

        if self.proc is not None:
            self.thread.join()
            try:
                if self.proc.wait() != os.EX_OK:
                    if not self.killed:
                        writemsg(colorize("BAD", f"GPKG external program failed."))
                        raise CompressorOperationFailed("decompression failed")
            finally:
                self.proc.stdout.close()


class checksum_helper:
    """
    Do checksum generation and GPG Signature generation and verification
    """

    SIGNING = 0
    VERIFY = 1

    def __init__(self, settings, gpg_operation=None, detached=True, signature=None):
        """
        settings         # portage settings
        gpg_operation    # either SIGNING or VERIFY
        signature        # GPG signature string used for GPG verify only
        """
        self.settings = settings
        self.gpg_operation = gpg_operation
        self.gpg_proc = None
        self.gpg_result = None
        self.gpg_output = None
        self.finished = False
        self.sign_file_path = None

        if (gpg_operation == checksum_helper.VERIFY) and (os.getuid() == 0):
            try:
                drop_user = self.settings.get("GPG_VERIFY_USER_DROP", "nobody")
                if drop_user == "":
                    self.uid = None
                else:
                    self.uid = pwd.getpwnam(drop_user).pw_uid
            except KeyError:
                writemsg(colorize("BAD", f"!!! Failed to find user {drop_user}."))
                raise

            try:
                drop_group = self.settings.get("GPG_VERIFY_GROUP_DROP", "nogroup")
                if drop_group == "":
                    self.gid = None
                else:
                    self.gid = grp.getgrnam(drop_group).gr_gid
            except KeyError:
                writemsg(colorize("BAD", f"!!! Failed to find group {drop_group}."))
                raise
        else:
            self.uid = None
            self.gid = None

        # Initialize the hash libs
        self.libs = {}
        for hash_name in MANIFEST2_HASH_DEFAULTS:
            self.libs[hash_name] = checksum.hashfunc_map[hash_name]._hashobject()

        # GPG
        env = self.settings.environ()
        if self.gpg_operation == checksum_helper.SIGNING:
            gpg_signing_base_command = self.settings.get(
                "BINPKG_GPG_SIGNING_BASE_COMMAND"
            )
            digest_algo = self.settings.get("BINPKG_GPG_SIGNING_DIGEST")
            gpg_home = self.settings.get("BINPKG_GPG_SIGNING_GPG_HOME")
            gpg_key = self.settings.get("BINPKG_GPG_SIGNING_KEY")

            if detached:
                gpg_detached = "--detach-sig"
            else:
                gpg_detached = "--clear-sign"

            if gpg_signing_base_command:
                gpg_signing_command = gpg_signing_base_command.replace(
                    "[PORTAGE_CONFIG]",
                    f"--homedir {gpg_home} "
                    f"--digest-algo {digest_algo} "
                    f"--local-user {gpg_key} "
                    f"{gpg_detached} "
                    "--batch --no-tty",
                )

                gpg_signing_command = shlex_split(
                    varexpand(gpg_signing_command, mydict=self.settings)
                )
                gpg_signing_command = [x for x in gpg_signing_command if x != ""]
                try:
                    env["GPG_TTY"] = os.ttyname(sys.stdout.fileno())
                except OSError:
                    pass
            else:
                raise CommandNotFound("GPG signing command is not set")

            self.gpg_proc = subprocess.Popen(
                gpg_signing_command,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                env=env,
            )

        elif self.gpg_operation == checksum_helper.VERIFY:
            if (signature is None) and (detached == True):
                raise MissingSignature("No signature provided")

            gpg_verify_base_command = self.settings.get(
                "BINPKG_GPG_VERIFY_BASE_COMMAND"
            )
            gpg_home = self.settings.get("BINPKG_GPG_VERIFY_GPG_HOME")

            if not gpg_verify_base_command:
                raise CommandNotFound("GPG verify command is not set")

            gpg_verify_command = gpg_verify_base_command.replace(
                "[PORTAGE_CONFIG]", f"--homedir {gpg_home} "
            )

            if detached:
                self.sign_file_fd, self.sign_file_path = tempfile.mkstemp(
                    ".sig", "portage-sign-"
                )

                gpg_verify_command = gpg_verify_command.replace(
                    "[SIGNATURE]", f"{self.sign_file_path} -"
                )

                # Create signature file and allow everyone read
                with open(self.sign_file_fd, "wb") as sign:
                    sign.write(signature)
                os.chmod(self.sign_file_path, 0o644)
            else:
                gpg_verify_command = gpg_verify_command.replace(
                    "[SIGNATURE]", "--output - -"
                )

            gpg_verify_command = shlex_split(
                varexpand(gpg_verify_command, mydict=self.settings)
            )
            gpg_verify_command = [x for x in gpg_verify_command if x != ""]

            self.gpg_proc = subprocess.Popen(
                gpg_verify_command,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                env=env,
                user=self.uid,
                group=self.gid,
            )

    def __del__(self):
        self.finish()

    def _check_gpg_status(self, gpg_status):
        """
        Check GPG status log for extra info.
        GPG will return OK even if the signature owner is not trusted.
        """
        good_signature = False
        trust_signature = False

        for l in gpg_status.splitlines():
            if l.startswith("[GNUPG:] GOODSIG"):
                good_signature = True

            if l.startswith("[GNUPG:] TRUST_ULTIMATE") or l.startswith(
                "[GNUPG:] TRUST_FULLY"
            ):
                trust_signature = True

        if (not good_signature) or (not trust_signature):
            writemsg(colorize("BAD", f"!!!\n{self.gpg_result.decode()}"))
            raise InvalidSignature("GPG verify failed")

    def update(self, data):
        """
        Write data to hash libs and GPG stdin.
        """
        for c in self.libs:
            self.libs[c].update(data)

        if self.gpg_proc is not None:
            self.gpg_proc.stdin.write(data)

    def finish(self):
        """
        Tell GPG file is EOF, and get results, then do clean up.
        """
        if self.finished:
            return

        if self.gpg_proc is not None:
            # Tell GPG EOF
            self.gpg_proc.stdin.close()

            return_code = self.gpg_proc.wait()

            if self.sign_file_path:
                os.remove(self.sign_file_path)

            self.finished = True

            self.gpg_result = self.gpg_proc.stderr.read()
            self.gpg_output = self.gpg_proc.stdout.read()
            self.gpg_proc.stdout.close()
            self.gpg_proc.stderr.close()

            if return_code == os.EX_OK:
                if self.gpg_operation == checksum_helper.VERIFY:
                    self._check_gpg_status(self.gpg_result.decode())
            else:
                writemsg(colorize("BAD", f"!!!\n{self.gpg_result.decode()}"))
                if self.gpg_operation == checksum_helper.SIGNING:
                    writemsg(colorize("BAD", self.gpg_output.decode()))
                    raise GPGException("GPG signing failed")
                elif self.gpg_operation == checksum_helper.VERIFY:
                    raise InvalidSignature("GPG verify failed")


class tar_safe_extract:
    """
    A safer version of tar extractall that doing sanity check.
    Note that this does not solve all security problems.
    """

    def __init__(self, tar: tarfile.TarFile, prefix: str = ""):
        """
        tar: an opened TarFile that ready to be read.
        prefix: a optional prefix for an inner directory should be considered
        as the root directory. e.g. "metadata" and "image".
        """
        self.tar = tar
        self.prefix = prefix
        self.closed = False
        self.file_list = []

    def extractall(self, dest_dir: str):
        """
        Extract all files to a temporary directory in the dest_dir, and move
        them to the dest_dir after sanity check.
        """
        if self.closed:
            raise OSError("Tar file is closed.")
        temp_dir = tempfile.TemporaryDirectory(dir=dest_dir)
        try:
            while True:
                member = self.tar.next()
                if member is None:
                    break
                if (member.name in self.file_list) or (
                    os.path.join(".", member.name) in self.file_list
                ):
                    writemsg(
                        colorize(
                            "BAD", f"Danger: duplicate files detected: {member.name}"
                        )
                    )
                    raise ValueError("Duplicate files detected.")
                if member.name.startswith("/"):
                    writemsg(
                        colorize(
                            "BAD", f"Danger: absolute path detected: {member.name}"
                        )
                    )
                    raise ValueError("Absolute path detected.")
                if member.name.startswith("../") or ("/../" in member.name):
                    writemsg(
                        colorize(
                            "BAD", f"Danger: path traversal detected: {member.name}"
                        )
                    )
                    raise ValueError("Path traversal detected.")
                if member.isdev():
                    writemsg(
                        colorize("BAD", f"Danger: device file detected: {member.name}")
                    )
                    raise ValueError("Device file detected.")
                if member.islnk() and (member.linkname not in self.file_list):
                    writemsg(
                        colorize(
                            "BAD", f"Danger: hardlink escape detected: {member.name}"
                        )
                    )
                    raise ValueError("Hardlink escape detected.")

                self.file_list.append(member.name)
                self.tar.extract(member, path=temp_dir.name)

            data_dir = os.path.join(temp_dir.name, self.prefix)
            for file in os.listdir(data_dir):
                shutil.move(os.path.join(data_dir, file), os.path.join(dest_dir, file))
        finally:
            temp_dir.cleanup()
            self.closed = True


class gpkg:
    """
    Gentoo binary package
    https://www.gentoo.org/glep/glep-0078.html
    """

    def __init__(self, settings, basename=None, gpkg_file=None):
        """
        gpkg class handle all gpkg operations for one package.
        basename is the package basename.
        gpkg_file should be exists file path for read or will create.
        """
        self.settings = settings
        self.gpkg_version = "gpkg-1"
        if gpkg_file is None:
            self.gpkg_file = None
        else:
            self.gpkg_file = _unicode_decode(
                gpkg_file, encoding=_encodings["fs"], errors="strict"
            )

        if basename is None:
            self.basename = None
        else:
            self.basename = basename.split("/", maxsplit=1)[-1]

        self.checksums = []
        self.manifest_old = []
        self.signature_exist = None
        self.prefix = None

        # Compression is the compression algorithm, if set to None will
        # not use compression.
        self.compression = self.settings.get("BINPKG_COMPRESS", None)
        if self.compression in ["", "none"]:
            self.compression = None

        # The create_signature is whether create signature for the package or not.
        if "binpkg-signing" in self.settings.features:
            self.create_signature = True
        else:
            self.create_signature = False

        # The request_signature is whether signature files are mandatory.
        # If set true, any missing signature file will cause reject processing.
        if "binpkg-request-signature" in self.settings.features:
            self.request_signature = True
        else:
            self.request_signature = False

        # The verify_signature is whether verify package signature or not.
        # In rare case user may want to ignore signature,
        # E.g. package with expired signature.
        if "binpkg-ignore-signature" in self.settings.features:
            self.verify_signature = False
        else:
            self.verify_signature = True

        self.ext_list = {
            "gzip": ".gz",
            "bzip2": ".bz2",
            "lz4": ".lz4",
            "lzip": ".lz",
            "lzop": ".lzo",
            "xz": ".xz",
            "zstd": ".zst",
        }

    def unpack_metadata(self, dest_dir=None):
        """
        Unpack metadata to dest_dir.
        If dest_dir is None, return files and values in dict.
        The dict key will be UTF-8, not bytes.
        """
        self._verify_binpkg(metadata_only=True)

        with tarfile.open(self.gpkg_file, "r") as container:
            metadata_tarinfo, metadata_comp = self._get_inner_tarinfo(
                container, "metadata"
            )

            with tar_stream_reader(
                container.extractfile(metadata_tarinfo),
                self._get_decompression_cmd(metadata_comp),
            ) as metadata_reader:
                metadata_tar = io.BytesIO(metadata_reader.read())

            with tarfile.open(mode="r:", fileobj=metadata_tar) as metadata:
                if dest_dir is None:
                    metadata_ = {
                        os.path.relpath(k.name, "metadata"): metadata.extractfile(
                            k
                        ).read()
                        for k in metadata.getmembers()
                    }
                else:
                    metadata_safe = tar_safe_extract(metadata, "metadata")
                    metadata_safe.extractall(dest_dir)
                    metadata_ = True
            metadata_tar.close()
        return metadata_

    def get_metadata(self, want=None):
        """
        get package metadata.
        if want is list, return all want key-values in dict
        if want is str, return the want key value
        """
        if want is None:
            return self.unpack_metadata()
        elif isinstance(want, str):
            metadata = self.unpack_metadata()
            metadata_want = metadata.get(want, None)
            return metadata_want
        else:
            metadata = self.unpack_metadata()
            metadata_want = {k: metadata.get(k, None) for k in want}
            return metadata_want

    def get_metadata_url(self, url, want=None):
        """
        Return the requested metadata from url gpkg.
        Default return all meta data.
        Use 'want' to get specific name from metadata.
        This method only support the correct package format.
        Wrong files order or incorrect basename will be considered invalid
        to reduce potential attacks.
        Only signature will be check if the signature file is the next file.
        Manifest will be ignored since it will be at the end of package.
        """
        # The init download file head size
        init_size = 51200

        # Load remote container
        container_file = io.BytesIO(
            urlopen(url, headers={"Range": "bytes=0-" + str(init_size)}).read()
        )

        # Check gpkg and metadata
        with tarfile.open(mode="r", fileobj=container_file) as container:
            if self.gpkg_version not in (
                os.path.basename(f) for f in container.getnames()
            ):
                raise InvalidBinaryPackageFormat("Invalid gpkg file.")

            metadata_tarinfo, metadata_comp = self._get_inner_tarinfo(
                container, "metadata"
            )

            # Extra 10240 bytes for signature
            end_size = metadata_tarinfo.offset_data + metadata_tarinfo.size + 10240
            _, remainder = divmod(end_size, 512)
            end_size += 512 - remainder

            # If need more data
            if end_size > 10000000:
                raise InvalidBinaryPackageFormat("metadata too large " + str(end_size))
            if end_size > init_size:
                container_file.seek(0, io.SEEK_END)
                container_file.write(
                    urlopen(
                        url,
                        headers={"Range": f"bytes={init_size + 1}-{end_size}"},
                    ).read()
                )

        container_file.seek(0)

        # Reload and process full metadata
        with tarfile.open(mode="r", fileobj=container_file) as container:
            metadata_tarinfo, metadata_comp = self._get_inner_tarinfo(
                container, "metadata"
            )

            # Verify metadata file signature if needed
            # binpkg-ignore-signature can override this.
            signature_filename = metadata_tarinfo.name + ".sig"
            if signature_filename in container.getnames():
                if self.request_signature and self.verify_signature:
                    metadata_signature = container.extractfile(
                        signature_filename
                    ).read()
                    checksum_info = checksum_helper(
                        self.settings,
                        gpg_operation=checksum_helper.VERIFY,
                        signature=metadata_signature,
                    )
                    checksum_info.update(container.extractfile(metadata_tarinfo).read())
                    checksum_info.finish()

            # Load metadata
            with tar_stream_reader(
                container.extractfile(metadata_tarinfo),
                self._get_decompression_cmd(metadata_comp),
            ) as metadata_reader:
                metadata_file = io.BytesIO(metadata_reader.read())

            with tarfile.open(mode="r:", fileobj=metadata_file) as metadata:
                if want is None:
                    metadata_ = {
                        os.path.relpath(k.name, "metadata"): metadata.extractfile(
                            k
                        ).read()
                        for k in metadata.getmembers()
                    }
                else:
                    metadata_ = {
                        os.path.relpath(k.name, "metadata"): metadata.extractfile(
                            k
                        ).read()
                        for k in metadata.getmembers()
                        if k in want
                    }
            metadata_file.close()
        container_file.close()
        return metadata_

    def compress(self, root_dir, metadata, clean=False):
        """
        Use initialized configuation create new gpkg file from root_dir.
        Will overwrite any exists file.
        metadata is a dict, the key will be file name, the value will be
        the file contents.
        """

        root_dir = normalize_path(
            _unicode_decode(root_dir, encoding=_encodings["fs"], errors="strict")
        )

        # Get pre image info
        container_tar_format, image_tar_format = self._get_tar_format_from_stats(
            *self._check_pre_image_files(root_dir)
        )

        # Long CPV
        if len(self.basename) >= 154:
            container_tar_format = tarfile.GNU_FORMAT

        # gpkg container
        container = tarfile.TarFile(
            name=self.gpkg_file, mode="w", format=container_tar_format
        )

        # gpkg version
        gpkg_version_file = tarfile.TarInfo(
            os.path.join(self.basename, self.gpkg_version)
        )
        gpkg_version_file.mtime = datetime.now().timestamp()
        container.addfile(gpkg_version_file)
        checksum_info = checksum_helper(self.settings)
        checksum_info.finish()
        self._record_checksum(checksum_info, gpkg_version_file)

        compression_cmd = self._get_compression_cmd()

        # metadata
        self._add_metadata(container, metadata, compression_cmd)

        # image
        if self.create_signature:
            checksum_info = checksum_helper(
                self.settings, gpg_operation=checksum_helper.SIGNING
            )
        else:
            checksum_info = checksum_helper(self.settings)

        image_tarinfo = self._create_tarinfo("image")
        image_tarinfo.mtime = datetime.now().timestamp()
        with tar_stream_writer(
            image_tarinfo, container, image_tar_format, compression_cmd, checksum_info
        ) as image_writer:
            with tarfile.open(
                mode="w|", fileobj=image_writer, format=image_tar_format
            ) as image_tar:
                image_tar.add(root_dir, "image", recursive=True)

        image_tarinfo = container.getmember(image_tarinfo.name)
        self._record_checksum(checksum_info, image_tarinfo)

        if self.create_signature:
            self._add_signature(checksum_info, image_tarinfo, container)

        self._add_manifest(container)

        # Check if all directories are the same in the container
        prefix = os.path.commonpath(container.getnames())
        if not prefix:
            raise InvalidBinaryPackageFormat(
                f"gpkg file structure mismatch in {self.gpkg_file}"
            )

        container.close()

    def decompress(self, decompress_dir):
        """
        decompress current gpkg to decompress_dir
        """
        decompress_dir = normalize_path(
            _unicode_decode(decompress_dir, encoding=_encodings["fs"], errors="strict")
        )

        self._verify_binpkg()
        os.makedirs(decompress_dir, mode=0o755, exist_ok=True)

        with tarfile.open(self.gpkg_file, "r") as container:
            image_tarinfo, image_comp = self._get_inner_tarinfo(container, "image")

            with tar_stream_reader(
                container.extractfile(image_tarinfo),
                self._get_decompression_cmd(image_comp),
            ) as image_tar:
                with tarfile.open(mode="r|", fileobj=image_tar) as image:
                    try:
                        image_safe = tar_safe_extract(image, "image")
                        image_safe.extractall(decompress_dir)
                    except Exception as ex:
                        writemsg(colorize("BAD", "!!!Extract failed."))
                        raise
                    finally:
                        image_tar.kill()

    def update_metadata(self, metadata, new_basename=None, force=False):
        """
        Update metadata in the gpkg file.
        """
        self._verify_binpkg()
        self.checksums = []
        if self.signature_exist and not force:
            raise SignedPackage("Cannot update a signed gpkg file")

        if new_basename is None:
            if self.basename:
                new_basename = self.basename
            elif self.prefix:
                new_basename = self.prefix
            else:
                raise InvalidBinaryPackageFormat("No basename or prefix specified")
        else:
            new_basename = new_basename.split("/", maxsplit=1)[-1]
            self.basename = new_basename

        with open(self.gpkg_file, "rb") as container:
            container_tar_format = self._get_tar_format(container)
            if container_tar_format is None:
                raise InvalidBinaryPackageFormat("Cannot identify tar format")

        # container
        tmp_gpkg_file_name = f"{self.gpkg_file}.{os.getpid()}"
        with tarfile.TarFile(
            name=tmp_gpkg_file_name, mode="w", format=container_tar_format
        ) as container:
            # gpkg version
            gpkg_version_file = tarfile.TarInfo(
                os.path.join(new_basename, self.gpkg_version)
            )
            gpkg_version_file.mtime = datetime.now().timestamp()
            container.addfile(gpkg_version_file)
            checksum_info = checksum_helper(self.settings)
            checksum_info.finish()
            self._record_checksum(checksum_info, gpkg_version_file)

            compression_cmd = self._get_compression_cmd()

            # metadata
            self._add_metadata(container, metadata, compression_cmd)

            # reuse other stuffs
            with tarfile.open(self.gpkg_file, "r") as container_old:
                manifest_old = self.manifest_old.copy()

                for m in manifest_old:
                    file_name_old = m[1]
                    if os.path.basename(file_name_old) == self.gpkg_version:
                        continue
                    if os.path.basename(file_name_old).startswith("metadata"):
                        continue
                    old_data_tarinfo = container_old.getmember(
                        os.path.join(self.prefix, file_name_old)
                    )
                    new_data_tarinfo = copy(old_data_tarinfo)
                    new_file_path = list(os.path.split(new_data_tarinfo.name))
                    new_file_path[0] = new_basename
                    new_data_tarinfo.name = os.path.join(*new_file_path)
                    container.addfile(
                        new_data_tarinfo, container_old.extractfile(old_data_tarinfo)
                    )
                    self.checksums.append(m)

            self._add_manifest(container)

            # Check if all directories are the same in the container
            prefix = os.path.commonpath(container.getnames())
            if not prefix:
                raise InvalidBinaryPackageFormat(
                    f"gpkg file structure mismatch in {self.gpkg_file}; files: "
                    f"{container.getnames()}"
                )

        shutil.move(tmp_gpkg_file_name, self.gpkg_file)

    def update_signature(self, keep_current_signature=False):
        """
        Add / update signature in the gpkg file.
        if keep_current_signature is True, keep the current signature, otherwise, re-signing it.
        """
        self.create_signature = True
        self._verify_binpkg()
        self.checksums = []

        with open(self.gpkg_file, "rb") as container:
            container_tar_format = self._get_tar_format(container)
            if container_tar_format is None:
                raise InvalidBinaryPackageFormat("Cannot identify tar format")

        # container
        tmp_gpkg_file_name = f"{self.gpkg_file}.{os.getpid()}"
        with tarfile.TarFile(
            name=tmp_gpkg_file_name, mode="w", format=container_tar_format
        ) as container:
            # gpkg version
            gpkg_version_file = tarfile.TarInfo(
                os.path.join(self.prefix, self.gpkg_version)
            )
            gpkg_version_file.mtime = datetime.now().timestamp()
            container.addfile(gpkg_version_file)
            checksum_info = checksum_helper(self.settings)
            checksum_info.finish()
            self._record_checksum(checksum_info, gpkg_version_file)

            # reuse other stuffs
            with tarfile.open(self.gpkg_file, "r") as container_old:
                manifest_old = self.manifest_old.copy()
                file_list_old = [f[1] for f in manifest_old]

                for m in manifest_old:
                    file_name_old = m[1]
                    if os.path.basename(file_name_old) == self.gpkg_version:
                        continue
                    if os.path.basename(file_name_old).endswith(".sig"):
                        continue
                    old_data_tarinfo = container_old.getmember(
                        os.path.join(self.prefix, file_name_old)
                    )
                    new_data_tarinfo = copy(old_data_tarinfo)

                    container.addfile(
                        new_data_tarinfo, container_old.extractfile(old_data_tarinfo)
                    )
                    self.checksums.append(m)

                    # Check if signature file exists and reuse or create new one.
                    if keep_current_signature and (
                        file_name_old + ".sig" in file_list_old
                    ):
                        old_data_sign_tarinfo = container_old.getmember(
                            file_name_old + ".sig"
                        )
                        new_data_sign_tarinfo = copy(old_data_sign_tarinfo)
                        container.addfile(
                            new_data_sign_tarinfo,
                            container_old.extractfile(old_data_sign_tarinfo),
                        )
                        for manifest_sign in manifest_old:
                            if manifest_sign[1] == file_name_old + ".sig":
                                self.checksums.append(manifest_sign)
                                break
                    else:
                        checksum_info = checksum_helper(
                            self.settings, gpg_operation=checksum_helper.SIGNING
                        )
                        checksum_info.update(
                            container_old.extractfile(old_data_tarinfo).read()
                        )
                        checksum_info.finish()
                        self._add_signature(checksum_info, new_data_tarinfo, container)

            self._add_manifest(container)

            # Check if all directories are the same in the container
            prefix = os.path.commonpath(container.getnames())
            if not prefix:
                raise InvalidBinaryPackageFormat(
                    f"gpkg file structure mismatch in {self.gpkg_file}"
                )

        shutil.move(tmp_gpkg_file_name, self.gpkg_file)

    def _add_metadata(self, container, metadata, compression_cmd):
        """
        add metadata to container
        """
        if metadata is None:
            metadata = {}
        metadata_tarinfo = self._create_tarinfo("metadata")
        metadata_tarinfo.mtime = datetime.now().timestamp()

        if self.create_signature:
            checksum_info = checksum_helper(
                self.settings, gpg_operation=checksum_helper.SIGNING
            )
        else:
            checksum_info = checksum_helper(self.settings)

        with tar_stream_writer(
            metadata_tarinfo,
            container,
            tarfile.USTAR_FORMAT,
            compression_cmd,
            checksum_info,
        ) as metadata_writer:
            with tarfile.open(
                mode="w|", fileobj=metadata_writer, format=tarfile.USTAR_FORMAT
            ) as metadata_tar:
                for m in metadata:
                    m_info = tarfile.TarInfo(os.path.join("metadata", m))
                    m_info.mtime = datetime.now().timestamp()

                    if isinstance(metadata[m], bytes):
                        m_data = io.BytesIO(metadata[m])
                    else:
                        m_data = io.BytesIO(metadata[m].encode("UTF-8"))

                    m_data.seek(0, io.SEEK_END)
                    m_info.size = m_data.tell()
                    m_data.seek(0)
                    metadata_tar.addfile(m_info, m_data)
                    m_data.close()

        metadata_tarinfo = container.getmember(metadata_tarinfo.name)
        self._record_checksum(checksum_info, metadata_tarinfo)

        if self.create_signature:
            self._add_signature(checksum_info, metadata_tarinfo, container)

    def _quickpkg(self, contents, metadata, root_dir, protect=None):
        """
        Similar to compress, but for quickpkg.
        Will compress the given files to image with root,
        ignoring all other files.
        """
        eout = EOutput()

        protect_file = io.BytesIO(
            b"# empty file because --include-config=n when `quickpkg` was used\n"
        )
        protect_file.seek(0, io.SEEK_END)
        protect_file_size = protect_file.tell()

        root_dir = normalize_path(
            _unicode_decode(root_dir, encoding=_encodings["fs"], errors="strict")
        )

        # Get pre image info
        container_tar_format, image_tar_format = self._get_tar_format_from_stats(
            *self._check_pre_quickpkg_files(contents, root_dir, ignore_missing=True)
        )

        # Long CPV
        if len(self.basename) >= 154:
            container_tar_format = tarfile.GNU_FORMAT

        # GPKG container
        container = tarfile.TarFile(
            name=self.gpkg_file, mode="w", format=container_tar_format
        )

        # GPKG version
        gpkg_version_file = tarfile.TarInfo(
            os.path.join(self.basename, self.gpkg_version)
        )
        gpkg_version_file.mtime = datetime.now().timestamp()
        container.addfile(gpkg_version_file)
        checksum_info = checksum_helper(self.settings)
        checksum_info.finish()
        self._record_checksum(checksum_info, gpkg_version_file)

        compression_cmd = self._get_compression_cmd()
        # Metadata
        self._add_metadata(container, metadata, compression_cmd)

        # Image
        if self.create_signature:
            checksum_info = checksum_helper(
                self.settings, gpg_operation=checksum_helper.SIGNING
            )
        else:
            checksum_info = checksum_helper(self.settings)

        paths = list(contents)
        paths.sort()
        image_tarinfo = self._create_tarinfo("image")
        image_tarinfo.mtime = datetime.now().timestamp()
        with tar_stream_writer(
            image_tarinfo, container, image_tar_format, compression_cmd, checksum_info
        ) as image_writer:
            with tarfile.open(
                mode="w|", fileobj=image_writer, format=image_tar_format
            ) as image_tar:
                if len(paths) == 0:
                    tarinfo = image_tar.tarinfo("image")
                    tarinfo.type = tarfile.DIRTYPE
                    tarinfo.size = 0
                    tarinfo.mode = 0o755
                    image_tar.addfile(tarinfo)

                for path in paths:
                    try:
                        lst = os.lstat(path)
                    except OSError as e:
                        if e.errno != errno.ENOENT:
                            raise
                        eout.ewarn(f'Missing file from local system: "{path}"')
                        del e
                        continue
                    contents_type = contents[path][0]
                    if path.startswith(root_dir):
                        arcname = "image/" + path[len(root_dir) :]
                    else:
                        raise ValueError(f"invalid root argument: '{root_dir}'")
                    live_path = path
                    if (
                        "dir" == contents_type
                        and not stat.S_ISDIR(lst.st_mode)
                        and os.path.isdir(live_path)
                    ):
                        # Even though this was a directory in the original ${D}, it exists
                        # as a symlink to a directory in the live filesystem.  It must be
                        # recorded as a real directory in the tar file to ensure that tar
                        # can properly extract it's children.
                        live_path = os.path.realpath(live_path)
                        lst = os.lstat(live_path)

                    # Since os.lstat() inside TarFile.gettarinfo() can trigger a
                    # UnicodeEncodeError when python has something other than utf_8
                    # return from sys.getfilesystemencoding() (as in bug #388773),
                    # we implement the needed functionality here, using the result
                    # of our successful lstat call. An alternative to this would be
                    # to pass in the fileobj argument to TarFile.gettarinfo(), so
                    # that it could use fstat instead of lstat. However, that would
                    # have the unwanted effect of dereferencing symlinks.

                    tarinfo = image_tar.tarinfo(arcname)
                    tarinfo.mode = lst.st_mode
                    tarinfo.uid = lst.st_uid
                    tarinfo.gid = lst.st_gid
                    tarinfo.size = 0
                    tarinfo.mtime = lst.st_mtime
                    tarinfo.linkname = ""
                    if stat.S_ISREG(lst.st_mode):
                        inode = (lst.st_ino, lst.st_dev)
                        if (
                            lst.st_nlink > 1
                            and inode in image_tar.inodes
                            and arcname != image_tar.inodes[inode]
                        ):
                            tarinfo.type = tarfile.LNKTYPE
                            tarinfo.linkname = image_tar.inodes[inode]
                        else:
                            image_tar.inodes[inode] = arcname
                            tarinfo.type = tarfile.REGTYPE
                            tarinfo.size = lst.st_size
                    elif stat.S_ISDIR(lst.st_mode):
                        tarinfo.type = tarfile.DIRTYPE
                    elif stat.S_ISLNK(lst.st_mode):
                        tarinfo.type = tarfile.SYMTYPE
                        tarinfo.linkname = os.readlink(live_path)
                    else:
                        continue
                    try:
                        tarinfo.uname = pwd.getpwuid(tarinfo.uid)[0]
                    except KeyError:
                        pass
                    try:
                        tarinfo.gname = grp.getgrgid(tarinfo.gid)[0]
                    except KeyError:
                        pass

                    if stat.S_ISREG(lst.st_mode):
                        if protect and protect(path):
                            protect_file.seek(0)
                            tarinfo.size = protect_file_size
                            image_tar.addfile(tarinfo, protect_file)
                        else:
                            path_bytes = _unicode_encode(
                                path, encoding=_encodings["fs"], errors="strict"
                            )

                            with open(path_bytes, "rb") as f:
                                image_tar.addfile(tarinfo, f)

                    else:
                        image_tar.addfile(tarinfo)

        image_tarinfo = container.getmember(image_tarinfo.name)
        self._record_checksum(checksum_info, image_tarinfo)

        if self.create_signature:
            self._add_signature(checksum_info, image_tarinfo, container)

        self._add_manifest(container)

        # Check if all directories are the same in the container
        prefix = os.path.commonpath(container.getnames())
        if not prefix:
            raise InvalidBinaryPackageFormat(
                f"gpkg file structure mismatch in {self.gpkg_file}"
            )

        container.close()

    def _record_checksum(self, checksum_info, tarinfo):
        """
        Record checksum result for the given file.
        Replace old checksum if already exists.
        """

        # Remove prefix directory from the filename
        file_name = os.path.basename(tarinfo.name)

        for c in self.checksums:
            if c[1] == file_name:
                self.checksums.remove(c)

        checksum_record = ["DATA", file_name, str(tarinfo.size)]

        for c in checksum_info.libs:
            checksum_record.append(c)
            checksum_record.append(checksum_info.libs[c].hexdigest())

        self.checksums.append(checksum_record)

    def _add_manifest(self, container):
        """
        Add Manifest to the container based on current checksums.
        Creare GPG signatue if needed.
        """
        manifest = io.BytesIO()

        for m in self.checksums:
            manifest.write((" ".join(m) + "\n").encode("UTF-8"))

        if self.create_signature:
            checksum_info = checksum_helper(
                self.settings, gpg_operation=checksum_helper.SIGNING, detached=False
            )
            checksum_info.update(manifest.getvalue())
            checksum_info.finish()
            manifest.seek(0)
            manifest.write(checksum_info.gpg_output)

        if self.basename is not None:
            basename = self.basename
        elif self.prefix is not None:
            basename = self.prefix
        else:
            raise InvalidBinaryPackageFormat("No basename or prefix specified")

        manifest_tarinfo = tarfile.TarInfo(os.path.join(basename, "Manifest"))
        manifest_tarinfo.size = manifest.tell()
        manifest_tarinfo.mtime = datetime.now().timestamp()
        manifest.seek(0)
        container.addfile(manifest_tarinfo, manifest)
        manifest.close()

    def _load_manifest(self, manifest_string):
        """
        Check, load, and return manifest in a list by files
        """
        manifest = []
        manifest_filenames = []

        for manifest_record in manifest_string.splitlines():
            if manifest_record == "":
                continue
            manifest_record = manifest_record.strip().split()

            if manifest_record[0] != "DATA":
                raise DigestException("invalied Manifest")

            if manifest_record[1] in manifest_filenames:
                raise DigestException("Manifest duplicate file exists")

            try:
                int(manifest_record[2])
            except ValueError:
                raise DigestException("Manifest invalied file size")

            manifest.append(manifest_record)
            manifest_filenames.append(manifest_record[1])

        return manifest

    def _add_signature(self, checksum_info, tarinfo, container, manifest=True):
        """
        Add GPG signature for the given tarinfo file.
        manifest: add to manifest
        """
        if checksum_info.gpg_output is None:
            raise GPGException("GPG signature is not exists")

        signature = io.BytesIO(checksum_info.gpg_output)
        signature_tarinfo = tarfile.TarInfo(f"{tarinfo.name}.sig")
        signature_tarinfo.size = len(signature.getvalue())
        signature_tarinfo.mtime = datetime.now().timestamp()
        container.addfile(signature_tarinfo, signature)

        if manifest:
            signature_checksum_info = checksum_helper(self.settings)
            signature.seek(0)
            signature_checksum_info.update(signature.read())
            signature_checksum_info.finish()
            self._record_checksum(signature_checksum_info, signature_tarinfo)

        signature.close()

    def _verify_binpkg(self, metadata_only=False):
        """
        Verify current GPKG file.
        """
        # Check file path
        if self.gpkg_file is None:
            raise FileNotFound("no gpkg file provided")

        # Check if is file
        if not os.path.isfile(self.gpkg_file):
            raise FileNotFound(f"File not found {self.gpkg_file}")

        # Check if is tar file
        with open(self.gpkg_file, "rb") as container:
            container_tar_format = self._get_tar_format(container)
            if container_tar_format is None:
                get_binpkg_format(self.gpkg_file, check_file=True)
                raise InvalidBinaryPackageFormat(
                    f"Cannot identify tar format: {self.gpkg_file}"
                )

        # Check container
        with tarfile.open(self.gpkg_file, "r") as container:
            try:
                container_files = container.getnames()
            except tarfile.ReadError:
                get_binpkg_format(self.gpkg_file, check_file=True)
                raise InvalidBinaryPackageFormat(
                    f"Cannot read tar file: {self.gpkg_file}"
                )

            # Check if gpkg version file exists in any place
            if self.gpkg_version not in (os.path.basename(f) for f in container_files):
                get_binpkg_format(self.gpkg_file, check_file=True)
                raise InvalidBinaryPackageFormat(f"Invalid gpkg file: {self.gpkg_file}")

            # Check how many layers are in the container
            for f in container_files:
                if f.startswith("/"):
                    raise InvalidBinaryPackageFormat(
                        f"gpkg file structure mismatch '{f}' in {self.gpkg_file}"
                    )
                if f.count("/") != 1:
                    raise InvalidBinaryPackageFormat(
                        f"gpkg file structure mismatch '{f}' in {self.gpkg_file}"
                    )

            # Check if all directories are the same in the container
            prefix = os.path.commonpath(container_files)
            if not prefix:
                raise InvalidBinaryPackageFormat(
                    f"gpkg file structure mismatch in {self.gpkg_file}, {str(container_files)}"
                )

            gpkg_version_file = os.path.join(prefix, self.gpkg_version)

            # If any signature exists, we assume all files have signature.
            if any(f.endswith(".sig") for f in container_files):
                signature_exist = True
            else:
                signature_exist = False

            # Check if all files are unique to avoid same name attack
            container_files_unique = []
            for f in container_files:
                if f in container_files_unique:
                    raise InvalidBinaryPackageFormat(
                        f"Duplicate file {f} exist, potential attack?"
                    )
                container_files_unique.append(f)

            del container_files_unique

            # Add all files to check list
            unverified_files = container_files.copy()

            # Check Manifest file
            manifest_filename = os.path.join(prefix, "Manifest")
            if manifest_filename not in unverified_files:
                raise MissingSignature(f"Manifest not found: {self.gpkg_file}")

            manifest_file = container.extractfile(manifest_filename)
            manifest_data = manifest_file.read()
            manifest_file.close()

            if b"-----BEGIN PGP SIGNATURE-----" in manifest_data:
                signature_exist = True

            # Check Manifest signature if needed.
            # binpkg-ignore-signature can override this.
            if self.request_signature or signature_exist:
                checksum_info = checksum_helper(
                    self.settings, gpg_operation=checksum_helper.VERIFY, detached=False
                )

                try:
                    checksum_info.update(manifest_data)
                    checksum_info.finish()
                except (InvalidSignature, MissingSignature):
                    if self.verify_signature:
                        raise

                manifest_data = checksum_info.gpg_output
                unverified_files.remove(manifest_filename)
            else:
                unverified_files.remove(manifest_filename)

            # Load manifest and create manifest check list
            manifest = self._load_manifest(manifest_data.decode("UTF-8"))
            unverified_manifest = manifest.copy()

            # Check all remaining files
            for f in unverified_files.copy():
                if f.endswith(".sig"):
                    f_signature = None
                else:
                    f_signature = f + ".sig"

                # Find current file manifest record
                manifest_record = None
                for m in manifest:
                    if m[1] == os.path.basename(f):
                        manifest_record = m

                if manifest_record is None:
                    raise DigestException(f"{f} checksum not found in {self.gpkg_file}")

                if int(manifest_record[2]) != int(container.getmember(f).size):
                    raise DigestException(
                        f"{f} file size mismatched in {self.gpkg_file}"
                    )

                # Ignore image file and signature if not needed
                if os.path.basename(f).startswith("image") and metadata_only:
                    unverified_files.remove(f)
                    unverified_manifest.remove(manifest_record)
                    continue

                # Verify current file signature if needed
                # binpkg-ignore-signature can override this.
                if (
                    (self.request_signature or signature_exist)
                    and self.verify_signature
                    and f_signature
                ):
                    if f_signature in unverified_files:
                        signature_file = container.extractfile(f_signature)
                        signature = signature_file.read()
                        signature_file.close()
                        checksum_info = checksum_helper(
                            self.settings,
                            gpg_operation=checksum_helper.VERIFY,
                            signature=signature,
                        )
                    elif f == gpkg_version_file:
                        # gpkg version file is not signed
                        checksum_info = checksum_helper(self.settings)
                    else:
                        raise MissingSignature(
                            f"{f} signature not found in {self.gpkg_file}"
                        )
                else:
                    checksum_info = checksum_helper(self.settings)

                # Verify current file checksum
                f_io = container.extractfile(f)
                while True:
                    buffer = f_io.read(HASHING_BLOCKSIZE)
                    if buffer:
                        checksum_info.update(buffer)
                    else:
                        checksum_info.finish()
                        break
                f_io.close()

                # At least one supported checksum must be checked
                verified_hash_count = 0
                for c in checksum_info.libs:
                    try:
                        if (
                            checksum_info.libs[c].hexdigest().lower()
                            == manifest_record[manifest_record.index(c) + 1].lower()
                        ):
                            verified_hash_count += 1
                        else:
                            raise DigestException(
                                f"{f} checksum mismatched in {self.gpkg_file}"
                            )
                    except KeyError:
                        # Checksum method not supported
                        pass

                if verified_hash_count < 1:
                    raise DigestException(
                        f"{f} no supported checksum found in {self.gpkg_file}"
                    )

                # Current file verified
                unverified_files.remove(f)
                unverified_manifest.remove(manifest_record)

        # Check if any file IN Manifest but NOT IN binary package
        if len(unverified_manifest) != 0:
            raise DigestException(
                f"Missing files: {str(unverified_manifest)} in {self.gpkg_file}"
            )

        # Check if any file NOT IN Manifest but IN binary package
        if len(unverified_files) != 0:
            raise DigestException(
                f"Unknown files exists: {str(unverified_files)} in {self.gpkg_file}"
            )

        # Save current Manifest for other operations.
        self.manifest_old = manifest.copy()
        self.signature_exist = signature_exist
        self.prefix = prefix

    def _generate_metadata_from_dir(self, metadata_dir):
        """
        read all files in metadata_dir and return as dict
        """
        metadata = {}
        metadata_dir = normalize_path(
            _unicode_decode(metadata_dir, encoding=_encodings["fs"], errors="strict")
        )
        for parent, dirs, files in os.walk(metadata_dir):
            for f in files:
                try:
                    f = _unicode_decode(f, encoding=_encodings["fs"], errors="strict")
                except UnicodeDecodeError:
                    continue
                with open(os.path.join(parent, f), "rb") as metafile:
                    metadata[f] = metafile.read()
        return metadata

    def _get_binary_cmd(self, compression, mode):
        """
        get command list from portage and try match compressor
        """
        if compression not in _compressors:
            raise InvalidCompressionMethod(compression)

        compressor = _compressors[compression]
        if mode not in compressor:
            raise InvalidCompressionMethod(f"{compression}: {mode}")

        if mode == "compress" and (
            self.settings.get(f"BINPKG_COMPRESS_FLAGS_{compression.upper()}", None)
            is not None
        ):
            compressor["compress"] = compressor["compress"].replace(
                "${BINPKG_COMPRESS_FLAGS}",
                f"${{BINPKG_COMPRESS_FLAGS_{compression.upper()}}}",
            )

        cmd = compressor[mode].replace(
            "{JOBS}", str(makeopts_to_job_count(self.settings.get("MAKEOPTS", "1")))
        )
        cmd = shlex_split(varexpand(cmd, mydict=self.settings))

        # Filter empty elements that make Popen fail
        cmd = [x for x in cmd if x != ""]

        if (not cmd) and ((mode + "_alt") in compressor):
            cmd = shlex_split(
                varexpand(compressor[mode + "_alt"], mydict=self.settings)
            )
            cmd = [x for x in cmd if x != ""]

        if not cmd:
            raise CompressorNotFound(compression)
        if not find_binary(cmd[0]):
            raise CompressorNotFound(cmd[0])

        return cmd

    def _get_compression_cmd(self, compression=None):
        """
        return compression command for Popen
        """
        if compression is None:
            compression = self.compression
        if compression is None:
            return None
        else:
            return self._get_binary_cmd(compression, "compress")

    def _get_decompression_cmd(self, compression=None):
        """
        return decompression command for Popen
        """
        if compression is None:
            compression = self.compression
        if compression is None:
            return None
        else:
            return self._get_binary_cmd(compression, "decompress")

    def _get_tar_format(self, fileobj):
        """
        Try to detect tar version
        """
        old_position = fileobj.tell()
        fileobj.seek(0x101)
        magic = fileobj.read(8)
        fileobj.seek(0x9C)
        typeflag = fileobj.read(1)
        fileobj.seek(old_position)

        if magic == b"ustar  \x00":
            return tarfile.GNU_FORMAT
        elif magic == b"ustar\x0000":
            if typeflag == b"x" or typeflag == b"g":
                return tarfile.PAX_FORMAT
            else:
                return tarfile.USTAR_FORMAT

        return None

    def _get_tar_format_from_stats(
        self,
        image_max_prefix_length,
        image_max_name_length,
        image_max_linkname_length,
        image_max_file_size,
        image_total_size,
    ):
        """
        Choose the corresponding tar format according to
        the image information
        """
        # Max possible size in UStar is 8 GiB (8589934591 bytes)
        # stored in 11 octets
        # Use 8000000000, just in case we need add something extra

        # Total size > 8 GiB, container need use GNU tar format
        if image_total_size < 8000000000:
            container_tar_format = tarfile.USTAR_FORMAT
        else:
            container_tar_format = tarfile.GNU_FORMAT

        # Image at least one file > 8 GiB, image need use GNU tar format
        if image_max_file_size < 8000000000:
            image_tar_format = tarfile.USTAR_FORMAT
        else:
            image_tar_format = tarfile.GNU_FORMAT

        # UStar support max 155 prefix length, 100 file name and 100 link name,
        # ends with \x00. If any exceeded, failback to GNU format.
        if image_max_prefix_length >= 155:
            image_tar_format = tarfile.GNU_FORMAT

        if image_max_name_length >= 100:
            image_tar_format = tarfile.GNU_FORMAT

        if image_max_linkname_length >= 100:
            image_tar_format = tarfile.GNU_FORMAT
        return container_tar_format, image_tar_format

    def _check_pre_image_files(self, root_dir, image_prefix="image"):
        """
        Check the pre image files size and path, return the longest
        path length, largest single file size, and total files size.
        """
        image_prefix_length = len(image_prefix) + 1
        root_dir = os.path.join(
            normalize_path(
                _unicode_decode(root_dir, encoding=_encodings["fs"], errors="strict")
            ),
            "",
        )
        root_dir_length = len(
            _unicode_encode(root_dir, encoding=_encodings["fs"], errors="strict")
        )

        image_max_prefix_length = 0
        image_max_name_length = 0
        image_max_link_length = 0
        image_max_file_size = 0
        image_total_size = 0

        for parent, dirs, files in os.walk(root_dir):
            if portage.utf8_mode:
                parent = os.fsencode(parent)
                dirs = [os.fsencode(value) for value in dirs]
                files = [os.fsencode(value) for value in files]

            parent = _unicode_decode(parent, encoding=_encodings["fs"], errors="strict")
            for d in dirs:
                try:
                    d = _unicode_decode(d, encoding=_encodings["fs"], errors="strict")
                except UnicodeDecodeError as err:
                    writemsg(colorize("BAD", f"\n*** {err}\n\n"), noiselevel=-1)
                    raise

                d = os.path.join(parent, d)
                prefix_length = (
                    len(_unicode_encode(d, encoding=_encodings["fs"], errors="strict"))
                    - root_dir_length
                    + image_prefix_length
                )

                if os.path.islink(d):
                    path_link = os.readlink(d)
                    path_link_length = len(
                        _unicode_encode(
                            path_link, encoding=_encodings["fs"], errors="strict"
                        )
                    )
                    image_max_link_length = max(image_max_link_length, path_link_length)

                image_max_prefix_length = max(image_max_prefix_length, prefix_length)

            for f in files:
                try:
                    f = _unicode_decode(f, encoding=_encodings["fs"], errors="strict")
                except UnicodeDecodeError as err:
                    writemsg(colorize("BAD", f"\n*** {err}\n\n"), noiselevel=-1)
                    raise

                filename_length = len(
                    _unicode_encode(f, encoding=_encodings["fs"], errors="strict")
                )
                image_max_name_length = max(image_max_name_length, filename_length)

                f = os.path.join(parent, f)
                path_length = (
                    len(_unicode_encode(f, encoding=_encodings["fs"], errors="strict"))
                    - root_dir_length
                    + image_prefix_length
                )

                file_stat = os.lstat(f)

                if os.path.islink(f):
                    path_link = os.readlink(f)
                    path_link_length = len(
                        os.fsencode(path_link)
                        if portage.utf8_mode
                        else _unicode_encode(
                            path_link, encoding=_encodings["fs"], errors="strict"
                        )
                    )
                elif file_stat.st_nlink > 1:
                    # Hardlink exists
                    path_link_length = path_length
                else:
                    path_link_length = 0

                image_max_link_length = max(image_max_link_length, path_link_length)

                try:
                    file_size = os.path.getsize(f)
                except FileNotFoundError:
                    # Ignore file not found if symlink to non-existing file
                    if os.path.islink(f):
                        continue
                    else:
                        raise
                image_total_size += file_size
                image_max_file_size = max(image_max_file_size, file_size)

        return (
            image_max_prefix_length,
            image_max_name_length,
            image_max_link_length,
            image_max_file_size,
            image_total_size,
        )

    def _check_pre_quickpkg_files(
        self, contents, root, image_prefix="image", ignore_missing=False
    ):
        """
        Check the pre quickpkg files size and path, return the longest
        path length, largest single file size, and total files size.
        """
        image_prefix_length = len(image_prefix) + 1
        root_dir = os.path.join(
            normalize_path(
                _unicode_decode(root, encoding=_encodings["fs"], errors="strict")
            ),
            "",
        )
        root_dir_length = len(
            _unicode_encode(root_dir, encoding=_encodings["fs"], errors="strict")
        )

        image_max_prefix_length = 0
        image_max_name_length = 0
        image_max_link_length = 0
        image_max_file_size = 0
        image_total_size = 0

        paths = list(contents)
        for path in paths:
            try:
                path = _unicode_decode(path, encoding=_encodings["fs"], errors="strict")
            except UnicodeDecodeError as err:
                writemsg(colorize("BAD", f"\n*** {err}\n\n"), noiselevel=-1)
                raise

            d, f = os.path.split(path)

            prefix_length = (
                len(_unicode_encode(d, encoding=_encodings["fs"], errors="strict"))
                - root_dir_length
                + image_prefix_length
            )
            image_max_prefix_length = max(image_max_prefix_length, prefix_length)

            filename_length = len(
                _unicode_encode(f, encoding=_encodings["fs"], errors="strict")
            )
            image_max_name_length = max(image_max_name_length, filename_length)

            path_length = (
                len(_unicode_encode(path, encoding=_encodings["fs"], errors="strict"))
                - root_dir_length
                + image_prefix_length
            )

            if not os.path.exists(path):
                if ignore_missing:
                    continue
                else:
                    raise FileNotFound(path)

            file_stat = os.lstat(path)

            if os.path.islink(path):
                path_link = os.readlink(path)
                path_link_length = len(
                    _unicode_encode(
                        path_link, encoding=_encodings["fs"], errors="strict"
                    )
                )
            elif file_stat.st_nlink > 1:
                # Hardlink exists
                path_link_length = path_length
            else:
                path_link_length = 0

            image_max_link_length = max(image_max_link_length, path_link_length)

            if os.path.isfile(path):
                try:
                    file_size = os.path.getsize(path)
                except FileNotFoundError:
                    # Ignore file not found if symlink to non-existing file
                    if os.path.islink(path):
                        continue
                    else:
                        raise
                image_total_size += file_size
                if file_size > image_max_file_size:
                    image_max_file_size = file_size

        return (
            image_max_prefix_length,
            image_max_name_length,
            image_max_link_length,
            image_max_file_size,
            image_total_size,
        )

    def _create_tarinfo(self, file_name):
        """
        Create new tarinfo for the new file
        """
        if self.compression is None:
            ext = ""
        elif self.compression in self.ext_list:
            ext = self.ext_list[self.compression]
        else:
            raise InvalidCompressionMethod(self.compression)

        if self.basename:
            basename = self.basename
        elif self.prefix:
            basename = self.prefix
        else:
            raise InvalidBinaryPackageFormat("No basename or prefix specified")
        data_tarinfo = tarfile.TarInfo(os.path.join(basename, file_name + ".tar" + ext))
        return data_tarinfo

    def _extract_filename_compression(self, file_name):
        """
        Extract the file basename and compression method
        """
        file_name = os.path.basename(file_name)
        if file_name.endswith(".tar"):
            return file_name[:-4], None

        for compression in self.ext_list:
            if file_name.endswith(".tar" + self.ext_list[compression]):
                return (
                    file_name[: -len(".tar" + self.ext_list[compression])],
                    compression,
                )

        raise InvalidCompressionMethod(file_name)

    def _get_inner_tarinfo(self, tar, file_name):
        """
        Get inner tarinfo from given container.
        Will try get file_name from correct basename first,
        if it fail, try any file that have same name as file_name, and
        return the first one.
        """
        if self.gpkg_version not in (os.path.basename(f) for f in tar.getnames()):
            raise InvalidBinaryPackageFormat(f"Invalid gpkg file")

        if self.basename and self.prefix and not self.prefix.startswith(self.basename):
            writemsg(
                colorize("WARN", f"Package basename mismatched, using {self.prefix}")
            )

        all_files = tar.getmembers()
        for f in all_files:
            try:
                f_name, f_comp = self._extract_filename_compression(f.name)
            except InvalidCompressionMethod:
                continue

            if f_name == file_name:
                return f, f_comp

        # Not found
        raise FileNotFound(f"File Not found: {file_name}")
